import itertools
import logging
import base64
from collections import defaultdict, deque
from functools import reduce
from operator import mul

from sqlalchemy.dialects import postgresql

from raco import algebra, expression, rules, scheme
from raco import types
from raco.algebra import Shuffle
from raco.algebra import convertcondition
from raco.backends import Language, Algebra
from raco.backends.sql.catalog import SQLCatalog, PostgresSQLFunctionProvider
from raco.catalog import Catalog
from raco.datastructure.UnionFind import UnionFind
from raco.expression import UnnamedAttributeRef
from raco.expression import WORKERID, COUNTALL
from raco.representation import RepresentationProperties
from raco.rules import distributed_group_by, check_partition_equality
from raco.expression import util
from urlparse import urlparse

LOGGER = logging.getLogger(__name__)


def scheme_to_schema(s):
    if s:
        names, descrs = zip(*s.asdict.items())
        names = ["%s" % n for n in names]
        types_ = [r[1] for r in descrs]
    else:
        names = []
        types_ = []
    return {"columnTypes": types_, "columnNames": names}


def compile_expr(op, child_scheme, state_scheme):
    ####
    # Put special handling at the top!
    ####
    if isinstance(op, expression.NumericLiteral):
        if isinstance(op.value, int):
            myria_type = types.LONG_TYPE
        elif isinstance(op.value, float):
            myria_type = types.DOUBLE_TYPE
        else:
            raise NotImplementedError("Compiling NumericLiteral {} of type {}"
                                      .format(op, type(op.value)))

        return {
            'type': 'CONSTANT',
            'value': str(op.value),
            'valueType': myria_type
        }
    elif isinstance(op, expression.StringLiteral):
        return {
            'type': 'CONSTANT',
            'value': str(op.value),
            'valueType': types.STRING_TYPE
        }
    elif isinstance(op, expression.BooleanLiteral):
        return {
            'type': 'CONSTANT',
            'value': bool(op.value),
            'valueType': types.BOOLEAN_TYPE
        }
    elif isinstance(op, expression.BlobLiteral):
        return{
            'type': 'CONSTANT',
            'value': base64.standard_b64encode(op.value),
            'valueType': types.BLOB_TYPE
        }
    elif isinstance(op, expression.StateRef):
        return {
            'type': 'STATE',
            'columnIdx': op.get_position(child_scheme, state_scheme)
        }
    elif isinstance(op, expression.AttributeRef):
        return {
            'type': 'VARIABLE',
            'columnIdx': op.get_position(child_scheme, state_scheme)
        }
    elif isinstance(op, expression.Case):
        # Convert n-ary case statements to binary, as expected by Myria
        op = op.to_binary()
        assert len(op.when_tuples) == 1

        if_expr = compile_expr(op.when_tuples[0][0], child_scheme,
                               state_scheme)
        then_expr = compile_expr(op.when_tuples[0][1], child_scheme,
                                 state_scheme)
        else_expr = compile_expr(op.else_expr, child_scheme, state_scheme)

        return {
            'type': 'CONDITION',
            'children': [if_expr, then_expr, else_expr]
        }
    elif isinstance(op, expression.CAST):
        return {
            'type': 'CAST',
            'left': compile_expr(op.input, child_scheme, state_scheme),
            'right': {
                'type': 'TYPE',
                'outputType': op._type
            }
        }
    elif isinstance(op, expression.SUBSTR):
        children = []
        op.operands[1] = expression.CAST(types.INT_TYPE, op.operands[1])
        op.operands[2] = expression.CAST(types.INT_TYPE, op.operands[2])
        for operand in op.operands:
            children.append(compile_expr(operand, child_scheme, state_scheme))
        return {
            'type': op.opname(),
            'children': children
        }
    elif isinstance(op, expression.PYUDF):
        return {
            'type': op.opname(),
            'name': op.name,
            'outputType': op.typ,
            'children': [compile_expr(arg, child_scheme, state_scheme)
                         for arg in op.arguments]
        }

    ####
    # Everything below here is compiled automatically
    ####
    elif isinstance(op, expression.UnaryOperator):
        return {
            'type': op.opname(),
            'operand': compile_expr(op.input, child_scheme, state_scheme)
        }
    elif isinstance(op, expression.BinaryOperator):
        return {
            'type': op.opname(),
            'left': compile_expr(op.left, child_scheme, state_scheme),
            'right': compile_expr(op.right, child_scheme, state_scheme)
        }
    elif isinstance(op, expression.ZeroaryOperator):
        return {
            'type': op.opname(),
        }
    elif isinstance(op, expression.NaryOperator):
        children = []
        for operand in op.operands:
            children.append(compile_expr(operand, child_scheme, state_scheme))
        return {
            'type': op.opname(),
            'children': children
        }
    raise NotImplementedError("Compiling expr of class %s" % op.__class__)


def compile_mapping(expr, child_scheme, state_scheme):
    output_name, root_op = expr
    return {
        'outputName': output_name,
        'rootExpressionOperator': compile_expr(root_op,
                                               child_scheme,
                                               state_scheme)
    }


class MyriaLanguage(Language):
    reusescans = False


class MyriaOperator(object):
    language = MyriaLanguage


def relation_key_to_json(relation_key):
    return {"userName": relation_key.user,
            "programName": relation_key.program,
            "relationName": relation_key.relation}


class MyriaScan(algebra.Scan, MyriaOperator):

    def compileme(self):
        return {
            "opType": "TableScan",
            "relationKey": relation_key_to_json(self.relation_key),
            "debroadcast": self._debroadcast
        }


class MyriaScanTemp(algebra.ScanTemp, MyriaOperator):

    def compileme(self):
        return {
            "opType": "TempTableScan",
            "table": self.name,
            "debroadcast": self._debroadcast
        }


class MyriaFileScan(algebra.FileScan, MyriaOperator):

    def compileme(self):
        if self.format == 'OPP':
            encoding = dict({
                "opType": "SeaFlowScan",
                "source": self.get_source(self.path)
            }, **self.options)

        elif self.format == 'TIPSY':
            encoding = {
                "opType": "TipsyFileScan",
                "tipsyFilename": self.path,
                "iorderFilename": self.path + ".iord",
                "grpFilename": self.path + (
                    ".grp" if "group" not in self.options
                    else ".{}.grp".format(self.options["group"]))
            }

        else:
            parsed_url = urlparse(self.path)
            if parsed_url.scheme == 's3':
                encoding = {
                    "opType": "CSVFragmentTupleSource",
                    "reader": dict({
                        "readerType": "CSV",
                        "schema": scheme_to_schema(self.scheme())
                    }, **self.options),
                    "source": {
                        "dataType": "S3",
                        "s3Uri": self.path
                    }
                }
            else:
                encoding = {
                    "opType": "TupleSource",
                    "reader": dict({
                        "readerType": "CSV",
                        "schema": scheme_to_schema(self.scheme())
                    }, **self.options),
                    "source": {
                        "dataType": "URI",
                        "uri": self.path
                    }
                }

        return encoding


class MyriaLimit(algebra.Limit, MyriaOperator):

    def compileme(self, inputid):
        return {
            "opType": "Limit",
            "argChild": inputid,
            "numTuples": self.count,
        }


class MyriaUnionAll(algebra.UnionAll, MyriaOperator):

    def compileme(self, *args):
        return {
            "opType": "UnionAll",
            "argChildren": args
        }


class MyriaDifference(algebra.Difference, MyriaOperator):

    def compileme(self, leftid, rightid):
        return {
            "opType": "Difference",
            "argChild1": leftid,
            "argChild2": rightid,
        }


class MyriaSingleton(algebra.SingletonRelation, MyriaOperator):

    def compileme(self):
        return {
            "opType": "Singleton",
        }


class MyriaEmptyRelation(algebra.EmptyRelation, MyriaOperator):

    def compileme(self):
        return {
            "opType": "Empty",
            'schema': scheme_to_schema(self.scheme())
        }


class MyriaSelect(algebra.Select, MyriaOperator):

    def compileme(self, inputid):
        pred = compile_expr(self.condition, self.scheme(), None)
        return {
            "opType": "Filter",
            "argChild": inputid,
            "argPredicate": {
                "rootExpressionOperator": pred
            }
        }


class MyriaCrossProduct(algebra.CrossProduct, MyriaOperator):

    def compileme(self, leftid, rightid):
        column_names = [name for (name, _) in self.scheme()]
        allleft = [i.position for i in self.left.scheme().ascolumnlist()]
        allright = [i.position for i in self.right.scheme().ascolumnlist()]
        return {
            "opType": "SymmetricHashJoin",
            "argColumnNames": column_names,
            "argChild1": leftid,
            "argChild2": rightid,
            "argColumns1": [],
            "argColumns2": [],
            "argSelect1": allleft,
            "argSelect2": allright
        }


class MyriaStore(algebra.Store, MyriaOperator):

    def compileme(self, inputid):
        distributeFunction = None
        attributes = self.partitioning().hash_partitioned
        if attributes:
            indexes = [attr.get_position(self.scheme()) for attr in attributes]
            distributeFunction = {
                "type": "Hash",
                "indexes": indexes
            }
        elif self.partitioning().broadcasted:
            distributeFunction = {
                "type": "Broadcast"
            }
        else:
            distributeFunction = {
                "type": "RoundRobin"
            }
        return {
            "opType": "DbInsert",
            "relationKey": relation_key_to_json(self.relation_key),
            "argOverwriteTable": True,
            "argChild": inputid,
            "distributeFunction": distributeFunction
        }


class MyriaStoreTemp(algebra.StoreTemp, MyriaOperator):

    def compileme(self, inputid):
        distributeFunction = None
        attributes = self.partitioning().hash_partitioned
        if attributes:
            indexes = [attr.get_position(self.scheme()) for attr in attributes]
            distributeFunction = {
                "type": "Hash",
                "indexes": indexes
            }
        elif self.partitioning().broadcasted:
            distributeFunction = {
                "type": "Broadcast"
            }
        else:
            distributeFunction = {
                "type": "RoundRobin"
            }
        return {
            "opType": "TempInsert",
            "table": self.name,
            "argOverwriteTable": True,
            "argChild": inputid,
            "distributeFunction": distributeFunction
        }


class MyriaSink(algebra.Sink, MyriaOperator):

    def compileme(self, inputid):
        return {
            "opType": "EmptySink",
            "argChild": inputid,
        }


class MyriaAppendTemp(algebra.AppendTemp, MyriaOperator):

    def compileme(self, inputid):
        return {
            "opType": "TempInsert",
            "table": self.name,
            "argOverwriteTable": False,
            "argChild": inputid,
        }


def convert_nary_conditions(conditions, schemes):
    """Convert an NaryJoin map from global column index to local"""
    attr_map = {}   # map of global attribute to local column index
    count = 0
    for i, scheme in enumerate(schemes):
        for j, attr in enumerate(scheme.ascolumnlist()):
            attr_map[count] = [i, j]
            count += 1
    new_conditions = []   # arrays of [child_index, column_index]
    for join_cond in conditions:
        new_join_cond = []
        for attr in join_cond:
            new_join_cond.append(attr_map[attr.position])
        new_conditions.append(new_join_cond)
    return new_conditions


class MyriaSymmetricHashJoin(algebra.ProjectingJoin, MyriaOperator):

    def compileme(self, leftid, rightid):
        """Compile the operator to a sequence of json operators"""

        left_len = len(self.left.scheme())
        combined = self.left.scheme() + self.right.scheme()
        leftcols, rightcols = convertcondition(self.condition,
                                               left_len,
                                               combined)

        if self.output_columns is None:
            self.output_columns = self.scheme().ascolumnlist()
        column_names = [name for (name, _) in self.scheme()]
        pos = [i.get_position(combined) for i in self.output_columns]
        side = [p >= left_len for p in pos]
        assert sorted(side) == side, \
            "MyriaSymmetricHashJoin always emits left columns first"
        allleft = [i for i in pos if i < left_len]
        allright = [i - left_len for i in pos if i >= left_len]

        join = {
            "opType": "SymmetricHashJoin",
            "argColumnNames": column_names,
            "argChild1": "%s" % leftid,
            "argColumns1": leftcols,
            "argChild2": "%s" % rightid,
            "argColumns2": rightcols,
            "argSelect1": allleft,
            "argSelect2": allright,
            "argOrder": self.pull_order_policy
        }

        return join


class MyriaIDBController(algebra.IDBController, MyriaOperator):

    def compileme(self, *args):
        ret = {
            "opType": "IDBController",
            "argSelfIdbId": "%s" % self.idb_id,
            "argInitialInput": "%s" % args[0],
            "argIterationInput": "%s" % args[1],
            "argEosControllerInput": "%s" % args[2],
            "argState": self.get_agg(),
            "sync": self.recursion_mode == "SYNC"
        }
        if self.relation_key is not None:
            ret.update({
                "relationKey": relation_key_to_json(self.relation_key),
            })
        return ret


class MyriaEOSController(algebra.EOSController, MyriaOperator):

    def compileme(self, input):
        return {
            "opType": "EOSController",
            "argChild": input
        }


class MyriaLeapFrogJoin(algebra.NaryJoin, MyriaOperator):

    def compileme(self, *args):
        def convert_join_cond(pos_to_rel_col, cond, scheme):
            join_col_pos = [c.get_position(scheme) for c in cond]
            return [pos_to_rel_col[p] for p in join_col_pos]
        # map a output column to its origin
        rel_of_pos = {}     # pos => [rel_idx, field_idx]
        schemes = [c.scheme().ascolumnlist() for c in self.children()]
        pos = 0
        combined = []
        for rel_idx, scheme in enumerate(schemes):
            combined.extend(scheme)
            for field_idx in xrange(len(scheme)):
                rel_of_pos[pos] = [rel_idx, field_idx]
                pos += 1
        # build column names
        if self.output_columns is None:
            self.output_columns = self.scheme().ascolumnlist()
        column_names = [name for (name, _) in self.scheme()]
        # get rel_idx and field_idx of select columns
        out_pos_list = [
            i.get_position(combined) for i in list(self.output_columns)]
        output_fields = [rel_of_pos[p] for p in out_pos_list]
        join_fields = [
            convert_join_cond(rel_of_pos, cond, combined)
            for cond in self.conditions]
        return {
            "opType": "LeapFrogJoin",
            "joinFieldMapping": join_fields,
            "argColumnNames": column_names,
            "outputFieldMapping": output_fields,
            "argChildren": args
        }


class MyriaGroupBy(algebra.GroupBy, MyriaOperator):

    @staticmethod
    def agg_mapping(agg_expr):
        """Maps a BuiltinAggregateExpression to a Myria string constant
        representing the corresponding aggregate operation."""
        if isinstance(agg_expr, expression.MAX):
            return "MAX"
        elif isinstance(agg_expr, expression.MIN):
            return "MIN"
        elif isinstance(agg_expr, expression.SUM):
            return "SUM"
        elif isinstance(agg_expr, expression.AVG):
            return "AVG"
        elif isinstance(agg_expr, expression.STDEV):
            return "STDEV"
        elif isinstance(agg_expr, expression.COUNTALL):
            return "COUNT"
        raise NotImplementedError("MyriaGroupBy.agg_mapping({})".format(
            type(agg_expr)))

    @staticmethod
    def compile_builtin_agg(agg, child_scheme):
        assert isinstance(agg, expression.BuiltinAggregateExpression)

        if isinstance(agg, expression.COUNTALL):
            # Treat COUNT(*) the same as COUNT($0) since we don't support NULL
            column = 0
        else:
            assert isinstance(agg, expression.UnaryOperator)
            column = expression.toUnnamed(agg.input, child_scheme).position
        return {"type": "Primitive",
                "aggOps": [MyriaGroupBy.agg_mapping(agg)],
                "column": column}

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        group_fields = [expression.toUnnamed(ref, child_scheme)
                        for ref in self.grouping_list]

        built_ins = [agg_expr for agg_expr in self.aggregate_list
                     if isinstance(agg_expr,
                                   expression.BuiltinAggregateExpression)]

        aggregators = [MyriaGroupBy.compile_builtin_agg(agg_expr, child_scheme)
                       for agg_expr in built_ins]

        assert all(aggregators[i] != aggregators[j]
                   for i in range(len(aggregators))
                   for j in range(len(aggregators))
                   if i < j)

        udas = [agg_expr for agg_expr in self.aggregate_list
                if isinstance(agg_expr, expression.UdaAggregateExpression)]
        assert len(udas) + len(built_ins) == len(self.aggregate_list)

        if udas:
            inits = [compile_mapping(e, None, None) for e in self.inits]
            updates = [compile_mapping(e, child_scheme, self.state_scheme)
                       for e in self.updaters]
            emitters = [compile_mapping(("uda{i}".format(i=i),
                                         e.input),
                                        None, self.state_scheme)
                        for i, e in enumerate(udas)]
            aggregators.append({
                "type": "UserDefined",
                "initializers": inits,
                "updaters": updates,
                "emitters": emitters
            })

        return {
            "argChild": inputid,
            "aggregators": aggregators,
            "opType": "Aggregate",
            "argGroupFields": [field.position for field in group_fields]
        }


class MyriaInMemoryOrderBy(algebra.OrderBy, MyriaOperator):

    def compileme(self, inputsym):
        return {
            "opType": "InMemoryOrderBy",
            "argChild": inputsym,
            "argSortColumns": self.sort_columns,
            "argAscending": self.ascending
        }


class MyriaShuffle(algebra.Shuffle, MyriaOperator):

    """Represents a simple shuffle operator"""

    def compileme(self, inputid):
        raise NotImplementedError('shouldn''t ever get here, should be turned into SP-SC pair')  # noqa


class MyriaCollect(algebra.Collect, MyriaOperator):

    """Represents a simple collect operator"""

    def compileme(self, inputid):
        raise NotImplementedError('shouldn''t ever get here, should be turned into CP-CC pair')  # noqa


class MyriaDupElim(algebra.Distinct, MyriaOperator):

    """Represents duplicate elimination"""

    def compileme(self, inputid):
        return {
            "opType": "DupElim",
            "argChild": inputid,
        }


class MyriaApply(algebra.Apply, MyriaOperator):

    """Represents a simple apply operator"""

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        emitters = [compile_mapping(x, child_scheme, None)
                    for x in self.emitters]
        return {
            'opType': 'Apply',
            'argChild': inputid,
            'emitExpressions': emitters
        }


class MyriaStatefulApply(algebra.StatefulApply, MyriaOperator):

    """Represents a stateful apply operator"""

    def compileme(self, inputid):
        child_scheme = self.input.scheme()
        state_scheme = self.state_scheme

        def comp_map(x):
            return compile_mapping(x, child_scheme, state_scheme)

        emitters = [comp_map(x) for x in self.emitters]
        inits = [comp_map(x) for x in self.inits]
        updaters = [comp_map(x) for x in self.updaters]
        return {
            'opType': 'StatefulApply',
            'argChild': inputid,
            'emitExpressions': emitters,
            'initializerExpressions': inits,
            'updaterExpressions': updaters
        }


class MyriaBroadcastProducer(algebra.UnaryOperator, MyriaOperator):

    """A Myria BroadcastProducer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return algebra.Broadcast(self.input).partitioning()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            "opType": "BroadcastProducer",
            "argChild": inputid,
        }


class MyriaBroadcastConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria BroadcastConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'BroadcastConsumer',
            'argOperatorId': inputid
        }


class MyriaSplitProducer(algebra.UnaryOperator, MyriaOperator):

    """A Myria SplitProducer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)
        self.consumers = []

    def append_consumer(self, consumer):
        self.consumers.append(consumer)

    def shortStr(self):
        return self.opname()

    def __repr__(self):
        return "{op}({inp!r})".format(op=self.opname(), inp=self.input)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def compileme(self, inputid):
        return {
            "opType": "LocalMultiwayProducer",
            "argChild": inputid
        }


class MyriaSplitConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria SplitConsumer"""

    def __init__(self, input):
        assert isinstance(input, MyriaSplitProducer)
        algebra.UnaryOperator.__init__(self, input)
        input.append_consumer(self)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def shortStr(self):
        return self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'LocalMultiwayConsumer',
            'argOperatorId': inputid
        }


class MyriaShuffleProducer(algebra.UnaryOperator, MyriaOperator):

    """A Myria ShuffleProducer"""

    def __init__(self, input, hash_columns, shuffle_type=None):
        algebra.UnaryOperator.__init__(self, input)
        # If no specified shuffle type, it's a hash.
        # TODO: add support for more types, do not use None for Hash
        if shuffle_type is None:
            shuffle_type = Shuffle.ShuffleType.Hash
        if shuffle_type == Shuffle.ShuffleType.Identity:
            assert len(hash_columns) == 1
        self.hash_columns = hash_columns
        self.shuffle_type = shuffle_type
        self.buffer_type = None

    def shortStr(self):
        if self.shuffle_type == Shuffle.ShuffleType.Identity:
            return "%s(%s)" % (self.opname(), self.hash_columns[0])
        if self.shuffle_type == Shuffle.ShuffleType.RoundRobin:
            return "%s" % self.opname()
        if self.shuffle_type == Shuffle.ShuffleType.Hash:
            hash_string = ','.join([str(x) for x in self.hash_columns])
            return "%s(h(%s))" % (self.opname(), hash_string)

    def __repr__(self):
        return "{op}({inp!r}, {hc!r}, {st!r})".format(
            op=self.opname(), inp=self.input, hc=self.hash_columns,
            st=self.shuffle_type)

    def partitioning(self):
        return Shuffle(
            self.input,
            self.hash_columns,
            self.shuffle_type).partitioning()

    def set_buffer_type(self, buffer_type):
        self.buffer_type = buffer_type

    def num_tuples(self):
        return self.input.num_tuples()

    def compileme(self, inputid):
        if self.shuffle_type == Shuffle.ShuffleType.Hash:
            df = {
                "type": "Hash",
                "indexes": [x.position for x in self.hash_columns]
            }
        elif self.shuffle_type == Shuffle.ShuffleType.Identity:
            df = {
                "type": "Identity",
                "index": self.hash_columns[0].position
            }
        elif self.shuffle_type == Shuffle.ShuffleType.RoundRobin:
            df = {
                "type": "RoundRobin"
            }
        else:
            # TODO: merge HyperCubeShuffleProducer
            raise ValueError("Invalid ShuffleType")

        ret = {
            "opType": "ShuffleProducer",
            "argChild": inputid,
            "distributeFunction": df
        }
        if self.buffer_type is not None:
            ret.update({"argBufferStateType": self.buffer_type})

        return ret


class MyriaShuffleConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria ShuffleConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s" % self.opname()

    def partitioning(self):
        return self.input.partitioning()

    def compileme(self, inputid):
        return {
            'opType': 'ShuffleConsumer',
            'argOperatorId': inputid
        }


class MyriaConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria Consumer. Used as child for IDBController and EOSController"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)
        self.set_stop_recursion()

    def __repr__(self):
        return "{op}({inp!r})".format(inp=self.input, op=self.opname())

    def partitioning(self):
        return self.input.partitioning()

    def num_tuples(self):
        return self.input.num_tuples()

    def shortStr(self):
        return "%s(%s)" % (self.opname(), self.scheme())

    def scheme(self):
        if isinstance(self.input, MyriaIDBController):
            return scheme.Scheme([("idbID", types.INT_TYPE),
                                  ("isDeltaEmpty", types.BOOLEAN_TYPE)])
        return scheme.Scheme()

    def compileme(self, inputid):
        return {
            'opType': 'Consumer',
            'argOperatorId': inputid
        }


class MyriaCollectProducer(algebra.UnaryOperator, MyriaOperator):

    """A Myria CollectProducer"""

    def __init__(self, input, server):
        algebra.UnaryOperator.__init__(self, input)
        self.server = server

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        # TODO: have a way to say it is on a specific worker
        return RepresentationProperties()

    def shortStr(self):
        return "%s(@%s)" % (self.opname(), self.server)

    def compileme(self, inputid):
        return {
            "opType": "CollectProducer",
            "argChild": inputid,
        }

    def __repr__(self):
        return "{op}({inp!r}, {svr!r})".format(op=self.opname(),
                                               inp=self.input,
                                               svr=self.server)


class MyriaCollectConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria CollectConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputid):
        return {
            'opType': 'CollectConsumer',
            'argOperatorId': inputid
        }


class MyriaHyperCubeShuffle(algebra.HyperCubeShuffle, MyriaOperator):

    """Represents a HyperCubeShuffle shuffle operator"""

    def compileme(self, inputsym):
        raise NotImplementedError('shouldn''t ever get here, should be turned into HCSP-HCSC pair')  # noqa


class MyriaHyperCubeShuffleProducer(algebra.UnaryOperator, MyriaOperator):

    """A Myria HyperCubeShuffleProducer"""

    def __init__(self, input, hashed_columns,
                 hyper_cube_dims, mapped_hc_dims, cell_partition):
        algebra.UnaryOperator.__init__(self, input)
        self.hashed_columns = hashed_columns
        self.mapped_hc_dimensions = mapped_hc_dims
        self.hyper_cube_dimensions = hyper_cube_dims
        self.cell_partition = cell_partition

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return MyriaHyperCubeShuffle(
            self.input,
            self.hashed_columns,
            self.mapped_hc_dims,
            self.hyper_cube_dimensions,
            self.cell_partition).partitioning()

    def shortStr(self):
        mapping = {i: '*' for i in range(len(self.hyper_cube_dimensions))}
        mapping.update({h: 'h({col})'.format(col=i)
                        for i, h in zip(self.hashed_columns,
                                        self.mapped_hc_dimensions)})
        hash_string = ','.join(s for m, s in sorted(mapping.items()))
        return "%s(%s)" % (self.opname(), hash_string)

    def compileme(self, inputsym):
        return {
            "opType": "HyperCubeShuffleProducer",
            "distributeFunction": {
                "hashedColumns": list(self.hashed_columns),
                "mappedHCDimensions": list(self.mapped_hc_dimensions),
                "hyperCubeDimensions": list(self.hyper_cube_dimensions),
                "cellPartition": self.cell_partition
            },
            "argChild": inputsym
        }


class MyriaHyperCubeShuffleConsumer(algebra.UnaryOperator, MyriaOperator):

    """A Myria HyperCubeShuffleConsumer"""

    def __init__(self, input):
        algebra.UnaryOperator.__init__(self, input)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return self.input.partitioning()

    def shortStr(self):
        return "%s" % self.opname()

    def compileme(self, inputsym):
        return {
            "opType": "HyperCubeShuffleConsumer",
            "argOperatorId": inputsym
        }


class MyriaQueryScan(algebra.ZeroaryOperator, MyriaOperator):

    """A Myria Query Scan"""

    def __init__(self, sql, scheme, source_relation_keys,
                 num_tuples=algebra.DEFAULT_CARDINALITY,
                 partitioning=RepresentationProperties(),
                 debroadcast=False):
        algebra.ZeroaryOperator.__init__(self)
        self.sql = str(sql)
        self._scheme = scheme
        self.source_relation_keys = source_relation_keys
        self._num_tuples = num_tuples
        self._partitioning = partitioning
        self._debroadcast = debroadcast

    def __repr__(self):
        return ("{op}({sql!r}, {sch!r}, {rels!r}, {nt!r}, {part!r}, {db!r})"
                .format(op=self.opname(), sql=self.sql, sch=self._scheme,
                        rels=self.source_relation_keys, nt=self._num_tuples,
                        part=self._partitioning, db=self._debroadcast))

    def num_tuples(self):
        return self._num_tuples

    def partitioning(self):
        return self._partitioning

    def shortStr(self):
        return "MyriaQueryScan({sql!r})".format(sql=self.sql)

    def scheme(self):
        return self._scheme

    def compileme(self):
        return {
            "opType": "DbQueryScan",
            "sql": self.sql,
            "schema": scheme_to_schema(self._scheme),
            "sourceRelationKeys": [
                relation_key_to_json(rk) for rk in self.source_relation_keys],
            "debroadcast": self._debroadcast
        }


class MyriaCalculateSamplingDistribution(algebra.UnaryOperator, MyriaOperator):

    """A Myria SamplingDistribution operator"""

    def __init__(self, input, sample_size, is_pct, sample_type):
        algebra.UnaryOperator.__init__(self, input)
        self.sample_size = sample_size
        self.is_pct = is_pct
        self.sample_type = sample_type

    def __repr__(self):
        return "{op}({inp!r}, {size!r}, {is_pct!r}, {type!r})".format(
            op=self.opname(),
            inp=self.input,
            size=self.sample_size,
            is_pct=self.is_pct,
            type=self.sample_type)

    def shortStr(self):
        pct = '%' if self.is_pct else ''
        return "{op}{type}({size}{pct})".format(op=self.opname(),
                                                type=self.sample_type,
                                                size=self.sample_size,
                                                pct=pct)

    def num_tuples(self):
        return self.input.num_tuples()

    def partitioning(self):
        return RepresentationProperties()

    def scheme(self):
        return self.input.scheme() + scheme.Scheme([('SampleSize',
                                                     types.LONG_TYPE), (
                                                    'SampleType',
                                                    self.sample_type)])

    def compileme(self, inputid):
        size_key = "samplePercentage" if self.is_pct else "sampleSize"
        return {
            "opType": "SamplingDistribution",
            "argChild": inputid,
            size_key: self.sample_size,
            "sampleType": self.sample_type
        }


class MyriaSample(algebra.BinaryOperator, MyriaOperator):

    """A Myria Sample operator"""

    def __init__(self, left, right, sample_size, is_pct, sample_type):
        algebra.BinaryOperator.__init__(self, left, right)
        # sample_size, sample_type, is_pct are just used for displaying.
        self.sample_size = sample_size
        self.is_pct = is_pct
        self.sample_type = sample_type

    def __repr__(self):
        return "{op}({l!r}, {r!r}, {size!r}, {is_pct!r}, {type!r})".format(
            op=self.opname(),
            l=self.left,
            r=self.right,
            size=self.sample_size,
            is_pct=self.is_pct,
            type=self.sample_type)

    def shortStr(self):
        pct = '%' if self.is_pct else ''
        return "{op}{type}({size}{pct})".format(op=self.opname(),
                                                type=self.sample_type,
                                                size=self.sample_size,
                                                pct=pct)

    def num_tuples(self):
        return self.sample_size

    def partitioning(self):
        return RepresentationProperties()

    def scheme(self):
        """The right operator is the one sampled from."""
        return self.right.scheme()

    def compileme(self, leftid, rightid):
        return {
            "opType": "Sample",
            "argChild1": leftid,
            "argChild2": rightid,
        }


class LogicalSampleToDistributedSample(rules.Rule):

    """Converts logical SampleScan to the sequence of physical operators."""

    def fire(self, expr):
        if isinstance(expr, algebra.SampleScan):
            samp_size = expr.sample_size
            is_pct = expr.is_pct
            samp_type = expr.sample_type
            # Each worker computes (WorkerID, LocalCount).
            scan_r = MyriaScan(expr.relation_key, expr.scheme())
            cnt_all = MyriaGroupBy(input=scan_r, aggregate_list=[COUNTALL()])
            apply_wid = MyriaApply([('WorkerID', WORKERID()),
                                    ('WorkerCount', UnnamedAttributeRef(0))],
                                   cnt_all)
            # Master collects the counts and generates a distribution.
            collect = MyriaCollect(apply_wid)
            samp_dist = MyriaCalculateSamplingDistribution(collect, samp_size,
                                                           is_pct, samp_type)
            # Master sends out how much each worker should sample.
            shuff = MyriaShuffle(samp_dist, [UnnamedAttributeRef(0)],
                                 Shuffle.ShuffleType.Identity)
            # Workers perform actual sampling.
            samp = MyriaSample(shuff, scan_r, samp_size, is_pct, samp_type)
            return samp
        else:
            return expr


class BreakShuffle(rules.Rule):

    def fire(self, expr):
        if not isinstance(expr, MyriaShuffle):
            return expr

        producer = MyriaShuffleProducer(expr.input, expr.columnlist,
                                        expr.shuffle_type)
        consumer = MyriaShuffleConsumer(producer)
        return consumer


class BreakHyperCubeShuffle(rules.Rule):

    def fire(self, expr):
        """
        self.hashed_columns = hashed_columns
        self.mapped_hc_dimensions = mapped_hc_dims
        self.hyper_cube_dimensions = hyper_cube_dims
        self.cell_partition = cell_partition
        """
        if not isinstance(expr, MyriaHyperCubeShuffle):
            return expr
        producer = MyriaHyperCubeShuffleProducer(
            expr.input, expr.hashed_columns, expr.hyper_cube_dimensions,
            expr.mapped_hc_dimensions, expr.cell_partition)
        consumer = MyriaHyperCubeShuffleConsumer(producer)
        return consumer


class BreakCollect(rules.Rule):

    def fire(self, expr):
        if not isinstance(expr, MyriaCollect):
            return expr

        producer = MyriaCollectProducer(expr.input, None)
        consumer = MyriaCollectConsumer(producer)
        return consumer


class BreakBroadcast(rules.Rule):

    def fire(self, expr):
        if not isinstance(expr, algebra.Broadcast):
            return expr

        producer = MyriaBroadcastProducer(expr.input)
        consumer = MyriaBroadcastConsumer(producer)
        return consumer


class BreakSplit(rules.Rule):

    def fire(self, expr):
        if not isinstance(expr, algebra.Split):
            return expr

        producer = MyriaSplitProducer(expr.input)
        consumer = MyriaSplitConsumer(producer)
        return consumer


class LimitOrderBy(rules.Rule):

    def fire(self, exp):
        if exp.__class__ == algebra.Limit and \
                isinstance(exp.children()[0], algebra.OrderBy):
            child = exp.children()[0]
            return MyriaLimit(exp.count, MyriaInMemoryOrderBy(
                algebra.Collect(
                    MyriaLimit(exp.count,
                               MyriaInMemoryOrderBy(child.input,
                                                    child.sort_columns,
                                                    child.ascending))),
                child.sort_columns, child.ascending))
        return exp


class ShuffleBeforeSetop(rules.Rule):

    def fire(self, exp):
        if not isinstance(exp, (algebra.Difference, algebra.Intersection)):
            return exp

        def shuffle_after(op):
            cols = [expression.UnnamedAttributeRef(i)
                    for i in range(len(op.scheme()))]

            if check_partition_equality(op, cols):
                return op
            else:
                return algebra.Shuffle(child=op, columnlist=cols)

        exp.left = shuffle_after(exp.left)
        exp.right = shuffle_after(exp.right)

        return exp

    def __str__(self):
        return "Setop => Shuffle(Setop)"


class ShuffleBeforeJoin(rules.Rule):

    def fire(self, expr):
        # If not a join, who cares?
        if not isinstance(expr, algebra.Join):
            return expr

        # Figure out which columns go in the shuffle
        left_cols, right_cols = \
            convertcondition(expr.condition,
                             len(expr.left.scheme()),
                             expr.left.scheme() + expr.right.scheme())

        # Left shuffle cols
        left_cols = [expression.UnnamedAttributeRef(i)
                     for i in left_cols]
        # Right shuffle cols
        right_cols = [expression.UnnamedAttributeRef(i)
                      for i in right_cols]

        if check_partition_equality(expr.left, left_cols):
            new_left = expr.left
        elif expr.left.partitioning().broadcasted:
            new_left = expr.left
        else:
            new_left = algebra.Shuffle(expr.left, left_cols)

        if check_partition_equality(expr.right, right_cols):
            new_right = expr.right
        elif expr.right.partitioning().broadcasted:
            new_right = expr.right
        else:
            new_right = algebra.Shuffle(expr.right, right_cols)

        # Construct the object!
        assert isinstance(expr, algebra.ProjectingJoin)
        if isinstance(expr, algebra.ProjectingJoin):
            return algebra.ProjectingJoin(expr.condition,
                                          new_left, new_right,
                                          expr.output_columns)

    def __str__(self):
        return "Join => Shuffle(Join)"


class ShuffleBeforeIDBController(rules.Rule):
    def fire(self, expr):
        if not isinstance(expr, algebra.IDBController):
            return expr
        group_list, agg = expr.get_group_agg()

        for idx in range(2):
            if not isinstance(expr.children()[idx],
                              (algebra.Shuffle, algebra.EmptyRelation)):
                expr.args[idx] = algebra.Shuffle(
                    expr.children()[idx],
                    [expression.UnnamedAttributeRef(i) for i in group_list])
        return expr


class HCShuffleBeforeNaryJoin(rules.Rule):

    def __init__(self, catalog):
        assert isinstance(catalog, Catalog)
        self.catalog = catalog
        super(HCShuffleBeforeNaryJoin, self).__init__()

    @staticmethod
    def reversed_index(child_schemes, conditions):
        """Return the reversed index of join conditions. The reverse index
           specify for each column on each relation, which hypercube dimension
           it is mapped to, -1 means this columns is not in the hyper cube
           (not joined).

        Keyword arguments:
        child_schemes -- schemes of children.
        conditions -- join conditions.
        """
        # make it -1 first
        r_index = [[-1] * len(scheme) for scheme in child_schemes]
        for i, jf_list in enumerate(conditions):
            for jf in jf_list:
                r_index[jf[0]][jf[1]] = i
        return r_index

    @staticmethod
    def workload(dim_sizes, child_sizes, r_index):
        """Compute the workload given a hyper cube size assignment"""
        load = 0.0
        for i, size in enumerate(child_sizes):
            # compute subcube sizes
            scale = 1
            for index in r_index[i]:
                if index != -1:
                    scale = scale * dim_sizes[index]
            # add load per server by child i
            load += float(child_sizes[i]) / float(scale)
        return load

    @staticmethod
    def get_hyper_cube_dim_size(num_server, child_sizes,
                                conditions, r_index):
        """Find the optimal hyper cube dimension sizes using BFS.

        Keyword arguments:
        num_server -- number of servers, this sets upper bound of HC cells.
        child_sizes -- cardinality of each child.
        child_schemes -- schemes of children.
        conditions -- join conditions.
        r_index -- reversed index of join conditions.
        """
        # Helper function: compute the product.
        def product(array):
            return reduce(mul, array, 1)
        # Use BFS to find the best possible assignment.
        this = HCShuffleBeforeNaryJoin
        visited = set()
        toVisit = deque()
        toVisit.append(tuple([1 for _ in conditions]))
        min_work_load = None
        opt_dim_sizes = [0] * len(conditions)
        while len(toVisit) > 0:
            dim_sizes = toVisit.pop()
            workload = this.workload(dim_sizes, child_sizes, r_index)
            if ((workload < min_work_load) or (
                workload == min_work_load and max(dim_sizes) < max(
                    opt_dim_sizes)) or (min_work_load is None)):
                min_work_load = this.workload(
                    dim_sizes, child_sizes, r_index)
                opt_dim_sizes = dim_sizes
            visited.add(dim_sizes)
            for i, d in enumerate(dim_sizes):
                new_dim_sizes = (dim_sizes[0:i] +
                                 tuple([dim_sizes[i] + 1]) +
                                 dim_sizes[i + 1:])
                if (product(new_dim_sizes) <= num_server and
                        new_dim_sizes not in visited):
                    toVisit.append(new_dim_sizes)
        return opt_dim_sizes, min_work_load

    @staticmethod
    def coord_to_worker_id(coordinate, dim_sizes):
        """Convert coordinate of cell to worker id

        Keyword arguments:
        coordinate -- coordinate of hyper cube cell.
        dim_sizes -- sizes of dimensons of hyper cube.
        """
        assert len(coordinate) == len(dim_sizes)
        ret = 0
        for k, v in enumerate(coordinate):
            ret += v * reduce(mul, dim_sizes[k + 1:], 1)
        return ret

    @staticmethod
    def get_cell_partition(dim_sizes, conditions,
                           child_schemes, child_idx, hashed_columns):
        """Generate the cell_partition for a specific child.

        Keyword arguments:
        dim_sizes -- size of each dimension of the hypercube.
        conditions -- each element is an array of (child_idx, column).
        child_schemes -- schemes of children.
        child_idx -- index of this child.
        hashed_columns -- hashed columns of this child.
        """
        assert len(dim_sizes) == len(conditions)
        # make life a little bit easier
        this = HCShuffleBeforeNaryJoin
        # get reverse index
        r_index = this.reversed_index(child_schemes, conditions)
        # find which dims in hyper cube this relation is involved
        hashed_dims = [r_index[child_idx][col] for col in hashed_columns]
        assert -1 not in hashed_dims
        # group by cell according to their projection on subcube voxel
        cell_partition = defaultdict(list)
        coor_ranges = [list(range(d)) for d in dim_sizes]
        for coordinate in itertools.product(*coor_ranges):
            # project a hypercube cell to a subcube voxel
            voxel = [coordinate[dim] for dim in hashed_dims]
            cell_partition[tuple(voxel)].append(
                this.coord_to_worker_id(coordinate, dim_sizes))
        return [wid for vox, wid in sorted(cell_partition.items())]

    def fire(self, expr):
        def add_hyper_shuffle():
            """ Helper function: put a HyperCube shuffle before each child."""
            # make calling static method easier
            this = HCShuffleBeforeNaryJoin
            # get child schemes
            child_schemes = [op.scheme() for op in expr.children()]
            # convert join conditions from expressions to 2d array
            conditions = convert_nary_conditions(
                expr.conditions, child_schemes)
            # get number of servers from catalog
            num_server = self.catalog.get_num_servers()
            # get estimated cardinalities of children
            child_sizes = [child.num_tuples() for child in expr.children()]
            # get reversed index of join conditions
            r_index = this.reversed_index(child_schemes, conditions)
            # compute optimal dimension sizes
            (dim_sizes, workload) = this.get_hyper_cube_dim_size(
                num_server, child_sizes, conditions, r_index)
            # specify HyperCube shuffle to each child
            new_children = []
            for child_idx, child in enumerate(expr.children()):
                # (mapped hc dimension, column index)
                hashed_fields = [(hc_dim, i)
                                 for i, hc_dim
                                 in enumerate(r_index[child_idx])
                                 if hc_dim != -1]
                mapped_dims, hashed_columns = zip(*sorted(hashed_fields))
                # get cell partition for child i
                cell_partition = this.get_cell_partition(
                    dim_sizes, conditions, child_schemes,
                    child_idx, hashed_columns)
                # generate new children
                new_children.append(
                    algebra.HyperCubeShuffle(
                        child, hashed_columns, mapped_dims,
                        dim_sizes, cell_partition))
            # replace the children
            expr.args = new_children

        # only apply to NaryJoin
        if not isinstance(expr, algebra.NaryJoin):
            return expr
        # check if HC shuffle has been placed before
        shuffled_child = [isinstance(op, algebra.HyperCubeShuffle)
                          for op in list(expr.children())]
        if all(shuffled_child):    # already shuffled
            assert len(expr.children()) > 0
            return expr
        elif any(shuffled_child):
            raise NotImplementedError("NaryJoin is partially shuffled?")
        else:                      # add shuffle and order by
            add_hyper_shuffle()
            return expr


class OrderByBeforeNaryJoin(rules.Rule):

    def fire(self, expr):
        # if not NaryJoin, who cares?
        if not isinstance(expr, algebra.NaryJoin):
            return expr
        ordered_child = sum(
            [1 for child in expr.children()
             if isinstance(child, algebra.OrderBy)])

        # already applied
        if ordered_child == len(expr.children()):
            return expr
        elif ordered_child > 0:
            raise Exception("children are partially ordered? ")

        new_children = []
        for child in expr.children():
            # check: this rule must be applied after shuffle
            assert isinstance(child, algebra.HyperCubeShuffle)
            ascending = [True] * len(child.hashed_columns)
            new_children.append(
                algebra.OrderBy(
                    child, child.hashed_columns, ascending))
        expr.args = new_children
        return expr


class BroadcastBeforeCross(rules.Rule):

    def fire(self, expr):
        # If not a CrossProduct, who cares?
        if not isinstance(expr, algebra.CrossProduct):
            return expr

        if (isinstance(expr.left, algebra.Broadcast) or
                isinstance(expr.right, algebra.Broadcast)):
            return expr

        try:
            # By default, broadcast the smaller child
            if expr.left.num_tuples() < expr.right.num_tuples():
                expr.left = algebra.Broadcast(expr.left)
            else:
                expr.right = algebra.Broadcast(expr.right)
        except NotImplementedError:
            # If cardinalities unknown, broadcast the right child
            expr.right = algebra.Broadcast(expr.right)

        return expr


class ShuffleAfterSingleton(rules.Rule):

    def fire(self, expr):
        if isinstance(expr, MyriaSingleton):
            return expr

        if isinstance(expr, algebra.SingletonRelation):
            return algebra.Shuffle(MyriaSingleton(), [UnnamedAttributeRef(0)])

        return expr


class ShuffleAfterFileScan(rules.Rule):

    def fire(self, expr):
        # don't shuffle any FileScan under an existing shuffle or broadcast
        if isinstance(expr, (algebra.Shuffle,
                             algebra.HyperCubeShuffle,
                             algebra.Broadcast)):
            for e in expr.walk():
                if isinstance(e, algebra.FileScan):
                    e._needs_shuffle = False
            return expr

        if isinstance(expr, algebra.FileScan):
            if expr._needs_shuffle:
                expr._needs_shuffle = False
                return algebra.Shuffle(
                    expr, shuffle_type=Shuffle.ShuffleType.RoundRobin)

        return expr

    def __str__(self):
        return "FileScan => Shuffle(FileScan)"


class PushSelectThroughShuffle(rules.Rule):

    """Push selections through a shuffle. We only need to push it just under
    the shuffle, since we can invoke the PushSelects rule again. We assume that
    PushSelects has already been invoked, so all selects that could be pushed
    through a shuffle are already immediately above the shuffle."""

    def fire(self, expr):
        if (isinstance(expr, algebra.Select) and
            isinstance(expr.input, (algebra.Shuffle,
                                    algebra.HyperCubeShuffle,
                                    algebra.Broadcast,
                                    algebra.Collect))):

            old_select = expr
            old_shuffle = expr.input
            shuffle_child = expr.input.input

            new_select = old_select.__class__()
            new_select.copy(old_select)
            new_select.input = shuffle_child

            new_shuffle = old_shuffle.__class__()
            new_shuffle.copy(old_shuffle)
            new_shuffle.input = new_select

            return new_shuffle

        return expr

    def __str__(self):
        return "Select(Shuffle) => Shuffle(Select)"


class AddAppendTemp(rules.Rule):

    def fire(self, op):
        if not isinstance(op, MyriaStoreTemp):
            return op

        child = op.input
        if not isinstance(child, MyriaUnionAll):
            return op

        # TODO: handle multiple children.
        if len(child.args) != 2:
            return op

        left = child.args[0]
        right = child.args[1]
        rel_name = op.name

        def is_scan(op):
            return (isinstance(op, MyriaScanTemp) and op.name == rel_name)
        if is_scan(left) and not any(is_scan(op) for op in right.walk()):
            return MyriaAppendTemp(name=rel_name, input=right)
        elif is_scan(right) and not any(is_scan(op) for op in left.walk()):
            return MyriaAppendTemp(name=rel_name, input=left)

        return op


class PushIntoSQL(rules.Rule):

    def __init__(self, dialect=None, push_grouping=False):
        self.dialect = dialect or postgresql.dialect()
        self.push_grouping = push_grouping
        super(PushIntoSQL, self).__init__()

    def fire(self, expr):
        if isinstance(expr, (algebra.Scan, algebra.ScanTemp)):
            return expr
        cat = SQLCatalog(provider=PostgresSQLFunctionProvider(),
                         push_grouping=self.push_grouping)
        try:
            scan_relations = [s.relation_key for s in expr.walk()
                              if isinstance(s, algebra.Scan)]
            has_debroadcast = any(isinstance(s, algebra.Scan) and
                                  s._debroadcast for s in expr.walk())
            sql_plan = cat.get_sql(expr)
            sql_string = sql_plan.compile(dialect=self.dialect)
            sql_string.visit_bindparam = sql_string.render_literal_bindparam
            return MyriaQueryScan(sql=sql_string.process(sql_plan),
                                  scheme=expr.scheme(),
                                  source_relation_keys=scan_relations,
                                  num_tuples=expr.num_tuples(),
                                  partitioning=expr.partitioning(),
                                  debroadcast=has_debroadcast)
        except NotImplementedError as e:
            LOGGER.warn("Error converting {plan}: {e}"
                        .format(plan=expr, e=e))
            return expr


class InsertSplit(rules.Rule):

    """Inserts an algebra.Split operator in every fragment that has multiple
    heavy-weight operators."""
    heavy_ops = (algebra.Store, algebra.StoreTemp,
                 algebra.CrossProduct, algebra.Join, algebra.NaryJoin,
                 algebra.GroupBy, algebra.OrderBy)

    def insert_split_before_heavy(self, op):
        """Walk the tree starting from op and insert a split when we
        encounter a heavyweight operator."""
        if isinstance(op, MyriaAlgebra.fragment_leaves):
            return op

        if isinstance(op, InsertSplit.heavy_ops):
            return algebra.Split(op)

        return op.apply(self.insert_split_before_heavy)

    def fire(self, op):
        if isinstance(op, InsertSplit.heavy_ops):
            return op.apply(self.insert_split_before_heavy)

        return op


class MergeToNaryJoin(rules.Rule):

    """Merge consecutive binary join into a single multiway join
    Note: this code assumes that the binary joins form a left deep tree
    before the merge."""
    @staticmethod
    def mergable(op):
        """Recursively checks whether an operator is mergable to NaryJoin.
        An operator will be merged to NaryJoin if its subtree contains
        only joins.
        """
        allowed_intermediate_types = (algebra.ProjectingJoin, algebra.Select)
        if issubclass(type(op), algebra.ZeroaryOperator):
            return True
        if not isinstance(op, allowed_intermediate_types):
            return False
        elif issubclass(type(op), algebra.UnaryOperator):
            return MergeToNaryJoin.mergable(op.input)
        elif issubclass(type(op), algebra.BinaryOperator):
            return (MergeToNaryJoin.mergable(op.left) and
                    MergeToNaryJoin.mergable(op.right))

    @staticmethod
    def collect_join_groups(op, conditions, children):
        assert isinstance(op, algebra.ProjectingJoin)
        assert (isinstance(op.right, algebra.Select) or
                issubclass(type(op.right), algebra.ZeroaryOperator))
        children.append(op.right)
        conjuncs = expression.extract_conjuncs(op.condition)
        for cond in conjuncs:
            conditions.get_or_insert(cond.left)
            conditions.get_or_insert(cond.right)
            conditions.union(cond.left, cond.right)
        scan_then_select = (isinstance(op.left, algebra.Select) and
                            isinstance(op.left.input, algebra.ZeroaryOperator))
        if (scan_then_select or
                issubclass(type(op.left), algebra.ZeroaryOperator)):
            children.append(op.left)
        else:
            assert isinstance(op.left, algebra.ProjectingJoin)
            MergeToNaryJoin.collect_join_groups(op.left, conditions, children)

    def fire(self, op):
        if not isinstance(op, algebra.ProjectingJoin):
            return op

        # if op is the only binary join, return
        if not isinstance(op.left, algebra.ProjectingJoin):
            return op
        # if it is not mergable, e.g. aggregation along the path, return
        if not MergeToNaryJoin.mergable(op):
            return op
        # do the actual merge
        # 1. collect join groups
        join_groups = UnionFind()
        children = []
        MergeToNaryJoin.collect_join_groups(
            op, join_groups, children)
        # 2. extract join groups from the union-find data structure
        join_conds = defaultdict(list)
        for field, key in join_groups.parents.items():
            join_conds[key].append(field)
        conditions = [v for (k, v) in join_conds.items()]
        # Note: a cost based join order optimization need to be implemented.
        ordered_conds = sorted(conditions, key=lambda cond: min(cond))
        # 3. reverse the children due to top-down tree traversal
        return algebra.NaryJoin(
            list(reversed(children)), ordered_conds, op.output_columns)


class GetCardinalities(rules.Rule):

    """ get cardinalities information of Zeroary operators.
    """

    def __init__(self, catalog):
        assert isinstance(catalog, Catalog)
        self.catalog = catalog
        super(GetCardinalities, self).__init__()

    def fire(self, expr):
        # if not Zeroary operator, who cares?
        if not issubclass(type(expr), algebra.ZeroaryOperator):
            return expr

        if issubclass(type(expr), algebra.Scan):
            rel = expr.relation_key
            expr._cardinality = self.catalog.num_tuples(rel)
            return expr
        expr._cardinality = algebra.DEFAULT_CARDINALITY
        return expr


# 6. shuffle logics, hyper_cube_shuffle_logic is only used in HCAlgebra
left_deep_tree_shuffle_logic = [
    ShuffleBeforeSetop(),
    ShuffleBeforeJoin(),
    ShuffleBeforeIDBController(),
    BroadcastBeforeCross(),
    ShuffleAfterSingleton(),
    LimitOrderBy(),
    ShuffleAfterFileScan(),
]


# 8. Myriafy logical operators
# replace logical operator with its corresponding Myria operators
myriafy = [
    rules.OneToOne(algebra.CrossProduct, MyriaCrossProduct),
    rules.OneToOne(algebra.Store, MyriaStore),
    rules.OneToOne(algebra.StoreTemp, MyriaStoreTemp),
    rules.OneToOne(algebra.StatefulApply, MyriaStatefulApply),
    rules.OneToOne(algebra.Apply, MyriaApply),
    rules.OneToOne(algebra.Select, MyriaSelect),
    rules.OneToOne(algebra.Distinct, MyriaDupElim),
    rules.OneToOne(algebra.Shuffle, MyriaShuffle),
    rules.OneToOne(algebra.HyperCubeShuffle, MyriaHyperCubeShuffle),
    rules.OneToOne(algebra.Collect, MyriaCollect),
    rules.OneToOne(algebra.ProjectingJoin, MyriaSymmetricHashJoin),
    rules.OneToOne(algebra.NaryJoin, MyriaLeapFrogJoin),
    rules.OneToOne(algebra.Scan, MyriaScan),
    rules.OneToOne(algebra.ScanTemp, MyriaScanTemp),
    rules.OneToOne(algebra.FileScan, MyriaFileScan),
    rules.OneToOne(algebra.SingletonRelation, MyriaSingleton),
    rules.OneToOne(algebra.EmptyRelation, MyriaEmptyRelation),
    rules.OneToOne(algebra.UnionAll, MyriaUnionAll),
    rules.OneToOne(algebra.Difference, MyriaDifference),
    rules.OneToOne(algebra.OrderBy, MyriaInMemoryOrderBy),
    rules.OneToOne(algebra.Limit, MyriaLimit),
    rules.OneToOne(algebra.Sink, MyriaSink),
    rules.OneToOne(algebra.IDBController, MyriaIDBController),
]

# 9. break communication boundary
# get producer/consumer pair
break_communication = [
    BreakHyperCubeShuffle(),
    BreakShuffle(),
    BreakCollect(),
    BreakBroadcast(),
    BreakSplit(),
]


class MyriaAlgebra(Algebra):

    """ Myria algebra abstract class"""
    language = MyriaLanguage

    fragment_leaves = (
        MyriaSplitConsumer,
        MyriaShuffleConsumer,
        MyriaCollectConsumer,
        MyriaBroadcastConsumer,
        MyriaHyperCubeShuffleConsumer,
        MyriaScan,
        MyriaScanTemp,
        MyriaFileScan,
        MyriaEmptyRelation,
        MyriaSingleton
    )


class FlattenUnionAll(rules.Rule):

    @staticmethod
    def collect_children(op):
        if isinstance(op, algebra.UnionAll):
            children = []
            for child in op.args:
                children += FlattenUnionAll.collect_children(child)
            return children
        return [op]

    def fire(self, op):
        if not isinstance(op, algebra.UnionAll):
            return op
        children = FlattenUnionAll.collect_children(op)
        if len(children) == 1:
            return children[0]
        return algebra.UnionAll(children)


class FillInJoinPullOrder(rules.Rule):

    def __init__(self):
        self._disabled = False

    def find_idb(self, op):
        for child in op.children():
            if isinstance(child, MyriaIDBController):
                return True
            if self.find_idb(child):
                return True
        return False

    def fire(self, op):
        if not isinstance(op, algebra.UntilConvergence):
            return op
        if op.pull_order_policy == 'ALTERNATE':
            return op
        for ch in op.walk():
            if isinstance(ch, MyriaSymmetricHashJoin):
                left_idb = self.find_idb(ch.left)
                right_idb = self.find_idb(ch.right)
                if op.pull_order_policy == 'PULL_IDB':
                    # ALTERNATE when both are IDBs
                    if left_idb and not right_idb:
                        ch.pull_order_policy = 'LEFT'
                    elif right_idb and not left_idb:
                        ch.pull_order_policy = 'RIGHT'
                elif op.pull_order_policy == 'PULL_EDB':
                    # ALTERNATE when both are EDBs
                    if left_idb and not right_idb:
                        ch.pull_order_policy = 'RIGHT'
                    elif right_idb and not left_idb:
                        ch.pull_order_policy = 'LEFT'
                elif op.pull_order_policy == 'BUILD_EDB':
                    # pick any EDB to save one hash table
                    if not left_idb:
                        ch.pull_order_policy = 'LEFT_EOS'
                    elif not right_idb:
                        ch.pull_order_policy = 'RIGHT_EOS'
        return op


class PropagateAsyncFTBuffer(rules.Rule):

    def __init__(self):
        self._disabled = False

    def fire(self, op):
        if not isinstance(op, MyriaIDBController):
            return op
        buffer_type = op.get_agg()
        if buffer_type['type'] == "CountFilter":
            # do not propagate CountFilter
            return op
        # only do it for the iterative input
        if isinstance(op.args[1], MyriaShuffleConsumer):
            op.args[1].input.set_buffer_type(buffer_type)
        return op


class StoreFromIDB(rules.Rule):

    @staticmethod
    def collect_and_replace(op, idbproducers):
        if not isinstance(op, algebra.NaryOperator):
            return
        if isinstance(op, algebra.UntilConvergence):
            for idbproducer in op.children():
                if isinstance(idbproducer.input, MyriaIDBController):
                    idbproducers[idbproducer.input.name] = idbproducer
            return
        newchildren = []
        for child in op.children():
            if isinstance(child, algebra.UntilConvergence):
                StoreFromIDB.collect_and_replace(child, idbproducers)
            if (isinstance(child, algebra.Store) and
                    isinstance(child.input, algebra.ScanIDB)):
                assert child.input.name in idbproducers
                idbproducers[child.input.name].input.relation_key = \
                    child.relation_key
            else:
                newchildren.append(child)
        op.args = newchildren

    def fire(self, op):
        self.collect_and_replace(op, {})
        return op

    def __str__(self):
        return ("Store[relation](IDBScan(IDBController)) -> "
                "IDBController[relation]")


def replace_child_with(op, child, replacement):
    if isinstance(op, algebra.UnaryOperator):
        op.input = replacement
    elif isinstance(op, algebra.BinaryOperator):
        if op.left == child:
            op.left = replacement
        else:
            op.right = replacement
    elif isinstance(op, algebra.NaryOperator):
        op.args[op.args.index(child)] = replacement
    else:  # should not happen
        assert False


class RemoveEmptyFilter(rules.Rule):

    def fire(self, op):
        for child in op.children():
            if isinstance(child, algebra.Select) and child.condition is None:
                replace_child_with(op, child, child.children()[0])
        return op

    def __str__(self):
        return "Select[cond=null](Input) -> Input"


class RemoveSingleSplit(rules.Rule):

    def fire(self, op):
        if not isinstance(op, algebra.UntilConvergence):
            return op

        parent_map = {}
        op.collectParents(parent_map)
        for idx, child in enumerate(op.children()):
            if not isinstance(child, MyriaSplitProducer):
                continue
            if len(child.consumers) != 1:
                continue
            producer = parent_map[id(child.consumers[0])][0]
            if not isinstance(producer, MyriaShuffleProducer):
                continue
            idb = child.children()[0]
            replace_child_with(producer, child.consumers[0], idb)
            if child.consumers[0].stop_recursion:
                parent_map[id(producer)][0].set_stop_recursion()
            op.args[idx] = producer
        return op

    def __str__(self):
        return ("ShuffleConsumer(ShuffleProducer(SplitConsumer(SplitProducer"
                "(input)))) -> ShuffleConsumer(ShuffleProducer(input))")


class DoUntilConvergence(rules.Rule):

    @staticmethod
    def replace_scan_with_idb_consumer(op, idb_controllers, idb_producers):
        for ch in op.children():
            if isinstance(ch, algebra.ScanIDB):
                if ch.name not in idb_producers:
                    idb_producers[ch.name] = MyriaSplitProducer(
                        idb_controllers[ch.name])
                consumer = MyriaSplitConsumer(idb_producers[ch.name])
                consumer.set_stop_recursion()
                replace_child_with(op, ch, consumer)
                if (isinstance(op, algebra.Select) and
                        (isinstance(op.condition, expression.boolean.GTEQ) or
                         isinstance(op.condition, expression.boolean.GT))):
                    group_list, agg = idb_controllers[ch.name].get_group_agg()
                    if isinstance(agg, expression.aggregate.COUNTALL):
                        if isinstance(op.condition, expression.boolean.GTEQ):
                            agg.threshold = op.condition.right.value
                        else:
                            # count generates integers, translate it to GTEQ
                            agg.threshold = op.condition.right.value + 1
                        op.condition = None
            elif ch.stop_recursion:
                pass
            else:
                DoUntilConvergence.replace_scan_with_idb_consumer(
                    ch, idb_controllers, idb_producers)

    def fire(self, op):
        if not isinstance(op, algebra.UntilConvergence):
            return op
        if any(isinstance(ch, MyriaEOSController) for ch in op.children()):
            return op

        eos_controller = MyriaEOSController()
        eos_consumers = []
        idb_controllers = {}
        for idb_controller in op.children():
            idb_controller.args[2] = MyriaConsumer(eos_controller)
            eos_consumers.append(MyriaConsumer(idb_controller))
            idb_controllers[idb_controller.name] = idb_controller
        if len(eos_consumers) == 1:
            eos_controller.input = eos_consumers[0]
        else:
            eos_controller.input = MyriaUnionAll(eos_consumers)

        idb_producers = {}
        for idb_controller in op.children():
            DoUntilConvergence.replace_scan_with_idb_consumer(
                idb_controller, idb_controllers, idb_producers)
        new_statements = [eos_controller]
        for idb_controller in op.children():
            if idb_controller.name in idb_producers:
                producer = idb_producers[idb_controller.name]
                producer.idx = len(new_statements)
                producer.parent = op
                new_statements.append(producer)
            else:
                new_statements.append(MyriaSink(idb_controller))
        op.args = new_statements
        return op

    def __str__(self):
        return ("DoUntilConvergence")


def idb_until_convergence(async_ft):
    ret = [DoUntilConvergence(), RemoveEmptyFilter(), StoreFromIDB(),
           RemoveSingleSplit(), FillInJoinPullOrder()]
    if async_ft is not None:
        ret.append(PropagateAsyncFTBuffer())
    return ret


class MyriaLeftDeepTreeAlgebra(MyriaAlgebra):

    """Myria physical algebra using left deep tree pipeline and 1-D shuffle"""

    def opt_rules(self, **kwargs):
        opt_grps_sequence = [
            rules.remove_trivial_sequences,
            [
                rules.SimpleGroupBy(),
                rules.CountToCountall(),  # TODO revisit when we have NULLs
                rules.ProjectToDistinctColumnSelect(),
                rules.DistinctToGroupBy(),
                rules.DedupGroupBy(),
            ],
            rules.push_select,
            rules.push_project,
            rules.push_apply,
            left_deep_tree_shuffle_logic,
            [PushSelectThroughShuffle()],
            rules.push_select,
            distributed_group_by(MyriaGroupBy),
            [rules.PushApply()],
            [LogicalSampleToDistributedSample()],
            [FlattenUnionAll()],
            [rules.DeDupBroadcastInputs()],
        ]

        if kwargs.get('push_sql', False):
            opt_grps_sequence.append([
                PushIntoSQL(dialect=kwargs.get('dialect'),
                            push_grouping=kwargs.get(
                                'push_sql_grouping', False))])

        compile_grps_sequence = [
            myriafy,
            [AddAppendTemp()],
            break_communication,
            idb_until_convergence(kwargs.get('async_ft')),
        ]

        if kwargs.get('add_splits', True):
            compile_grps_sequence.append([InsertSplit()])
        # Even when false, plans may already include (manually added) Splits,
        # so we always need BreakSplit
        compile_grps_sequence.append([BreakSplit()])

        rule_grps_sequence = opt_grps_sequence + compile_grps_sequence

        # flatten the rules lists
        rule_list = list(itertools.chain(*rule_grps_sequence))

        # disable specified rules
        rules.Rule.apply_disable_flags(rule_list, *kwargs.keys())

        return rule_list


class MyriaHyperCubeAlgebra(MyriaAlgebra):

    """Myria physical algebra using HyperCubeShuffle and LeapFrogJoin"""

    def opt_rules(self, **kwargs):
        # this rule is hyper cube shuffle specific
        merge_to_nary_join = [
            MergeToNaryJoin()
        ]

        # catalog aware hc shuffle rules, so put them here
        hyper_cube_shuffle_logic = [
            GetCardinalities(self.catalog),
            HCShuffleBeforeNaryJoin(self.catalog),
            OrderByBeforeNaryJoin(),
        ]

        opt_grps_sequence = [
            rules.remove_trivial_sequences,
            [
                rules.SimpleGroupBy(),
                # TODO revisit when we have NULL support.
                rules.CountToCountall(),
                rules.DistinctToGroupBy(),
                rules.DedupGroupBy(),
            ],
            rules.push_select,
            rules.push_project,
            merge_to_nary_join,
            rules.push_apply,
            left_deep_tree_shuffle_logic,
            [PushSelectThroughShuffle()],
            rules.push_select,
            distributed_group_by(MyriaGroupBy),
            [rules.DeDupBroadcastInputs()],
            hyper_cube_shuffle_logic
        ]

        if kwargs.get('push_sql', False):
            opt_grps_sequence.append([PushIntoSQL()])

        compile_grps_sequence = [
            myriafy,
            [AddAppendTemp()],
            break_communication
        ]

        if kwargs.get('add_splits', True):
            compile_grps_sequence.append([InsertSplit()])
        # Even when false, plans may already include (manually added) Splits,
        # so we always need BreakSplit
        compile_grps_sequence.append([BreakSplit()])

        rule_grps_sequence = opt_grps_sequence + compile_grps_sequence

        # flatten the rules lists
        rule_list = list(itertools.chain(*rule_grps_sequence))

        # disable specified rules
        rules.Rule.apply_disable_flags(rule_list, *kwargs.keys())

        return rule_list

    def __init__(self, catalog=None):
        self.catalog = catalog


class OpIdFactory(object):

    def __init__(self):
        self.count = 0

    def alloc(self):
        ret = self.count
        self.count += 1
        return ret

    def getter(self):
        return lambda: self.alloc()


def ensure_store_temp(label, op):
    """Returns a StoreTemp that assigns the given operator to a temp relation
    with the given name. Op must be a 'normal operator', i.e.,
    not a Store/StoreTemp or a control-flow sub-plan operator."""
    assert not isinstance(op, (algebra.Store, algebra.StoreTemp))
    assert not isinstance(op, (algebra.Sequence, algebra.Parallel,
                               algebra.DoWhile))
    assert isinstance(label, basestring) and len(label) > 0
    return MyriaStoreTemp(input=op, name=label)


def compile_fragment(frag_root, op_ids):
    """Given a root operator, produce a SubQueryEncoding."""

    def one_fragment(rootOp):
        """Given an operator that is the root of a query fragment/plan, extract
        the operators in the fragment. Assembles a list cur_frag of the
        operators in the current fragment, in preorder from the root.

        This operator also assembles a queue of the discovered roots of later
        fragments, e.g., when there is a ShuffleProducer below. The list of
        operators that should be treated as fragment leaves is given by
        MyriaAlgebra.fragment_leaves. """

        # The current fragment starts with the current root
        cur_frag = [rootOp]
        # Initially, there are no new roots discovered below leaves of this
        # fragment.
        queue = []
        if rootOp.stop_recursion:
            pass
        elif isinstance(rootOp, MyriaAlgebra.fragment_leaves):
            # The current root operator is a fragment leaf, such as a
            # ShuffleProducer. Append its children to the queue of new roots.
            for child in rootOp.children():
                queue.append(child)
        else:
            # Otherwise, the children belong in this fragment. Recursively go
            # discover their fragments, including the queue of roots below
            # their children.
            for child in rootOp.children():
                (child_frag, child_queue) = one_fragment(child)
                # Add their fragment onto this fragment
                cur_frag += child_frag
                # Add their roots-of-next-fragments into our queue
                queue += child_queue
        return (cur_frag, queue)

    def fragments(rootOp):
        """Given the root of a query plan, recursively determine all the
        fragments in it."""
        # The queue of fragment roots. Initially, just the root of this query
        queue = [rootOp]
        ret = []
        while len(queue) > 0:
            # Get the next fragment root
            rootOp = queue.pop(0)
            # .. recursively learn the entire fragment, and any newly
            # discovered roots.
            (op_frag, op_queue) = one_fragment(rootOp)
            # .. Myria JSON expects the fragment operators in reverse order,
            # i.e., root at the bottom.
            ret.append(list(reversed(op_frag)))
            # .. and collect the newly discovered fragment roots.
            queue.extend(op_queue)
        return ret

    def call_compile_me(op, op_ids):
        "A shortcut to call the operator's compile_me function."
        op_id = op_ids[id(op)]
        child_op_ids = [op_ids[id(child)] for child in op.children()]
        op_dict = op.compileme(*child_op_ids)
        op_dict['opName'] = op.shortStr()
        assert isinstance(op_id, int), (type(op_id), op_id)
        op_dict['opId'] = op_id
        return op_dict

    # Determine and encode the fragments.
    results = []
    for frag in fragments(frag_root):
        frag_compilated = {
            'operators': [call_compile_me(op, op_ids) for op in frag]}
        results.append(frag_compilated)

    return results


def compile_plan(plan_op):
    """Given a root operator in MyriaX physical algebra,
    produce the dictionary encoding of the physical plan, in other words, a
    nested collection of Java QueryPlan operators."""
    # A dictionary mapping each object to a unique, object-dependent id.
    # Since we want this to be truly unique for each object instance, even if
    # two objects are equal, we use id(obj) as the key.
    op_ids = defaultdict(OpIdFactory().getter())

    subplan_ops = (algebra.Parallel, algebra.Sequence, algebra.DoWhile,
                   algebra.UntilConvergence)
    if not isinstance(plan_op, subplan_ops):
        plan_op = algebra.Parallel([plan_op])

    if isinstance(plan_op, algebra.Parallel):
        frag_list = [compile_fragment(op, op_ids) for op in plan_op.children()]
        return {"type": "SubQuery",
                "fragments": list(itertools.chain(*frag_list))}

    elif isinstance(plan_op, algebra.Sequence):
        plan_list = [compile_plan(pl_op) for pl_op in plan_op.children()]
        return {"type": "Sequence", "plans": plan_list}

    elif isinstance(plan_op, algebra.DoWhile):
        children = plan_op.children()
        if len(children) < 2:
            raise ValueError('DoWhile must have at >= 2 children: body and condition')  # noqa
        condition = children[-1]
        if isinstance(condition, subplan_ops):
            raise ValueError('DoWhile condition cannot be a subplan op {cls}'.format(cls=condition.__class__))  # noqa
        condition = ensure_store_temp('__dowhile_{}_condition'.format(id(
            plan_op)), condition)
        plan_op.args = children[:-1] + [condition]
        body = [compile_plan(pl_op) for pl_op in plan_op.children()]
        return {"type": "DoWhile",
                "body": body,
                "condition": condition.name}

    elif isinstance(plan_op, algebra.UntilConvergence):

        frag_list = [compile_fragment(op, op_ids) for op in plan_op.children()]
        return {"type": "SubQuery",
                "fragments": list(itertools.chain(*frag_list))}

    raise NotImplementedError("compiling subplan op {}".format(type(plan_op)))


def compile_to_json(raw_query, logical_plan, physical_plan,
                    language="not specified", **kwargs):
    """This function compiles a physical query plan to the JSON suitable for
    submission to the Myria REST API server. The logical plan is converted to a
    string and passed along unchanged."""

    # Store/StoreTemp is a reasonable physical plan... for now.
    root_ops = (algebra.Store, algebra.StoreTemp, algebra.Sink)
    if isinstance(physical_plan, root_ops):
        physical_plan = algebra.Parallel([physical_plan])

    subplan_ops = (algebra.Parallel, algebra.Sequence, algebra.DoWhile,
                   algebra.UntilConvergence)
    assert isinstance(physical_plan, subplan_ops), \
        'Physical plan must be a subplan operator, not {}'.format(type(physical_plan))  # noqa

    # raw_query must be a string
    if not isinstance(raw_query, basestring):
        raise ValueError("raw query must be a string")

    return {"rawQuery": raw_query, "logicalRa": str(logical_plan),
            "language": language, "plan": compile_plan(physical_plan),
            "ftMode": kwargs.get('async_ft', "NONE")}



from raco.catalog import Catalog
import raco.scheme as scheme
from raco.representation import RepresentationProperties
from raco.expression import UnnamedAttributeRef as AttIndex
from raco.catalog import DEFAULT_CARDINALITY
from .errors import MyriaError


class MyriaCatalog(Catalog):

    def __init__(self, connection):
        self.connection = connection

    def get_scheme(self, rel_key):
        relation_args = {
            'userName': rel_key.user,
            'programName': rel_key.program,
            'relationName': rel_key.relation
        }
        if not self.connection:
            raise RuntimeError(
                "no schema for relation %s because no connection" % rel_key)
        try:
            dataset_info = self.connection.dataset(relation_args)
        except MyriaError:
            raise ValueError('No relation {} in the catalog'.format(rel_key))
        schema = dataset_info['schema']
        return scheme.Scheme(zip(schema['columnNames'], schema['columnTypes']))

    def get_num_servers(self):
        if not self.connection:
            raise RuntimeError("no connection.")
        return len(self.connection.workers_alive())

    def get_function(self, name):
        """ Get user defined function metadata """
        if not self.connection:
            raise RuntimeError("no connection.")

        try:
            function_info = self.connection.get_function(name)
        except MyriaError:
            raise ValueError("Function does not exist.")

        return function_info

    def num_tuples(self, rel_key):
        relation_args = {
            'userName': rel_key.user,
            'programName': rel_key.program,
            'relationName': rel_key.relation
        }
        if not self.connection:
            raise RuntimeError(
                "no cardinality of %s because no connection" % rel_key)
        try:
            dataset_info = self.connection.dataset(relation_args)
        except MyriaError:
            raise ValueError(rel_key)
        num_tuples = dataset_info['numTuples']
        assert isinstance(num_tuples, (int, long)), type(num_tuples)
        # that's a work round. numTuples is -1 if the dataset is old
        if num_tuples != -1:
            assert num_tuples >= 0
            return num_tuples
        return DEFAULT_CARDINALITY

    def partitioning(self, rel_key):
        relation_args = {
            'userName': rel_key.user,
            'programName': rel_key.program,
            'relationName': rel_key.relation
        }
        if not self.connection:
            raise RuntimeError(
                "no schema for relation %s because no connection" % rel_key)
        try:
            dataset_info = self.connection.dataset(relation_args)
        except MyriaError:
            raise ValueError('No relation {} in the catalog'.format(rel_key))
        distribute_function = dataset_info['howDistributed']['df']
        # TODO: can we do anything useful with other distribute functions?
        if distribute_function:
            if distribute_function['type'] == "Broadcast":
                return RepresentationProperties(broadcasted=True)
            if distribute_function['type'] == "Identity":
                index = distribute_function['index']
                return RepresentationProperties(
                    hash_partitioned=tuple(AttIndex(index)))
            elif distribute_function['type'] == "Hash":
                indexes = distribute_function['indexes']
                return RepresentationProperties(
                    hash_partitioned=tuple(AttIndex(i) for i in indexes))
        return RepresentationProperties()



# coding=utf-8
""" Exceptions that occur during Python->RACO conversion """


class PythonConvertException(Exception):
    """ Base class for conversion exceptions """
    pass


class PythonTokenException(PythonConvertException):
    """ Base exception class for errors associated with a specific token """
    def __init__(self, token, line, column):
        self.token = token
        self.line = line
        self.column = column


class PythonUnrecognizedTokenException(PythonTokenException):
    """ Error occurring when an unrecognized token is encountered """
    def __str__(self):
        return 'Conversion error at token %s on line %d:%d' % \
               (self.token, self.line, self.column)


class PythonOutOfRangeException(PythonUnrecognizedTokenException):
    """ Error occurring when a slice is out of range"""
    def __str__(self):
        return 'Slice out of range error near token %s on line %d:%d' % \
               (self.token, self.line, self.column)


class PythonSyntaxException(PythonConvertException):
    """ Error occurring when a Python source string contains a syntax error """
    def __init__(self, message, line, column):
        self.token = message
        self.line = line
        self.column = column

    def __str__(self):
        return 'Syntax error: %s (%d%s)' % \
               (self.token, self.line,
                ':' + str(self.column) if self.column else '')


class PythonUnsupportedOperationException(PythonSyntaxException):
    """ Error occurring when an unsupported operation is detected """
    def __str__(self):
        return 'Unsupported operation: %s (%d%s)' % \
               (self.token, self.line,
                ':' + str(self.column) if self.column else '')


class PythonArgumentException(PythonSyntaxException):
    """ Error occurring when a problem with an argument is detected """
    def __str__(self):
        return '%s (%d%s)' % \
               (self.token, self.line,
                ':' + str(self.column) if self.column else '')



import raco.myrial.groupby as groupby
import raco.myrial.multiway as multiway
from raco.myrial.cfg import ControlFlowGraph
from raco.myrial.emitarg import FullWildcardEmitArg, TableWildcardEmitArg
from raco.myrial.exceptions import *
import raco.algebra
import raco.expression
import raco.catalog
import raco.scheme
from raco.backends.myria import (MyriaLeftDeepTreeAlgebra,
                                 MyriaHyperCubeAlgebra,
                                 compile_to_json)
from raco.compile import optimize
from raco import relation_key
from raco.algebra import Shuffle

import collections
import copy
from functools import reduce


class DuplicateAliasException(Exception):

    """Bag comprehension arguments must have different alias names."""
    pass


class InvalidStatementException(Exception):
    pass


def get_unnamed_ref(column_ref, scheme, offset=0):
    """Convert a string or int into an attribute ref on the new table"""  # noqa
    if isinstance(column_ref, int):
        index = column_ref
    else:
        index = scheme.getPosition(column_ref)
    return raco.expression.UnnamedAttributeRef(index + offset)


def check_binop_compatability(op_name, left, right):
    """Check whether the arguments to an operation are compatible."""
    # Todo: check for type compatibility here?
    # https://github.com/uwescience/raco/issues/213
    if len(left.scheme()) != len(right.scheme()):
        raise SchemaMismatchException(op_name)


def check_assignment_compatability(before, after):
    """Check whether multiple assignments are compatible."""
    # TODO: check for exact schema match -- this is blocked by a general
    # cleanup of raco types.
    check_binop_compatability("assignment", before, after)


class ExpressionProcessor(object):

    """Convert syntactic expressions into relational algebra operations."""

    def __init__(self, symbols, catalog, use_dummy_schema=False):
        self.symbols = symbols
        self.catalog = catalog
        self.use_dummy_schema = use_dummy_schema

        # Variables accessed by the current operation
        self.uses_set = set()

    def get_and_clear_uses_set(self):
        """Retrieve the uses set and then clear its value."""
        try:
            return self.uses_set
        finally:
            self.uses_set = set()

    def evaluate(self, expr):
        method = getattr(self, expr[0].lower())
        return method(*expr[1:])

    def __lookup_symbol(self, _id):
        if _id not in self.symbols:
            raise NoSuchRelationException(_id)

        self.uses_set.add(_id)
        return copy.deepcopy(self.symbols[_id])

    def alias(self, _id):
        return self.__lookup_symbol(_id)

    def _get_scan_scheme(self, rel_key):
        try:
            return self.catalog.get_scheme(rel_key)
        except KeyError:
            if not self.use_dummy_schema:
                raise NoSuchRelationException(rel_key)
            # Create a dummy schema suitable for emitting plans
            return raco.scheme.DummyScheme()

    def scan(self, rel_key):
        """Scan a database table."""
        assert isinstance(rel_key, relation_key.RelationKey)
        scheme = self._get_scan_scheme(rel_key)
        return raco.algebra.Scan(rel_key, scheme,
                                 self.catalog.num_tuples(rel_key),
                                 self.catalog.partitioning(rel_key))

    def samplescan(self, rel_key, samp_size, is_pct, samp_type):
        """Sample a base relation."""
        assert isinstance(rel_key, relation_key.RelationKey)
        if samp_type not in ('WR', 'WoR'):
            raise MyrialCompileException(
                "Invalid Sampling Type: %s" % samp_type)
        scheme = self._get_scan_scheme(rel_key)
        return raco.algebra.SampleScan(rel_key, scheme, samp_size, is_pct,
                                       samp_type)

    def load(self, path, format, scheme, options):
        return raco.algebra.FileScan(path, format, scheme, options)

    def table(self, emit_clause):
        """Emit a single-row table literal."""
        emit_args = []
        for clause in emit_clause:
            emit_args.extend(clause.expand({}))

        from_args = collections.OrderedDict()
        from_args['$$SINGLETON$$'] = raco.algebra.SingletonRelation()

        # Add unbox relations to the from_args dictionary
        for name, sexpr in emit_args:
            self.extract_unbox_args(from_args, sexpr)

        op, info = multiway.merge(from_args)

        # rewrite clauses in terms of the new schema
        emit_args = [(name, multiway.rewrite_refs(sexpr, from_args, info))
                     for (name, sexpr) in emit_args]

        return raco.algebra.Apply(emitters=emit_args, input=op)

    @staticmethod
    def empty(_scheme):
        if not _scheme:
            _scheme = raco.scheme.Scheme()
        return raco.algebra.EmptyRelation(_scheme)

    def select(self, args):
        """Evaluate a select-from-where expression."""
        op = self.bagcomp(args.from_, args.where,
                          args.select, args.orderby, args.limit)
        if args.distinct:
            op = raco.algebra.Distinct(input=op)

        return op

    def extract_unbox_args(self, from_args, sexpr):
        """Extract unbox arguments from a scalar expression.

        :param from_args: An ordered dictionary that maps from a
        relation alias (string) to an instance of raco.algebra.Operator.
        :param sexpr: A scalar expression (raco.expression.Expresssion)
        instance.
        """
        for sub_expr in sexpr.walk():
            if isinstance(sub_expr, raco.expression.Unbox):
                name = sub_expr.table_name
                assert isinstance(name, basestring)
                if name not in from_args:
                    from_args[name] = self.__lookup_symbol(name)

    def bagcomp(self, from_clause, where_clause, emit_clause,
                orderby_clause, limit_clause):
        """Evaluate a bag comprehension.

        from_clause: A list of tuples of the form (id, expr).  expr can
        be None, which means "read the value from the symbol table".

        where_clause: An optional scalar expression (raco.expression).

        emit_clause: A list of EmitArg instances, each defining one or more
        output columns.

        orderby_clause: An optional list of OrderbyArg instances.

        limit_clause: An optional integer expression
        """

        # Make sure no aliases were reused: [FROM X, X EMIT *] is illegal
        from_aliases = set([x[0] for x in from_clause])
        if len(from_aliases) != len(from_clause):
            raise DuplicateAliasException()

        # For each FROM argument, create a mapping from ID to operator
        # (id, raco.algebra.Operator)
        from_args = collections.OrderedDict()

        for _id, expr in from_clause:
            assert isinstance(_id, basestring)
            if expr:
                from_args[_id] = self.evaluate(expr)
            else:
                from_args[_id] = self.__lookup_symbol(_id)

        # Expand wildcards into a list of output columns
        assert emit_clause  # There should always be something to emit
        emit_args = []
        statemods = []
        for clause in emit_clause:
            emit_args.extend(clause.expand(from_args))
            statemods.extend(clause.get_statemods())

        orig_op, _info = multiway.merge(from_args)
        orig_schema_length = len(orig_op.scheme())

        # Add unbox relations to the from_args dictionary
        for name, sexpr in emit_args:
            self.extract_unbox_args(from_args, sexpr)
        if where_clause:
            self.extract_unbox_args(from_args, where_clause)

        # Create a single RA operation that is the cross of all targets
        op, info = multiway.merge(from_args)

        # HACK: calculate unboxed columns as implicit grouping columns,
        # so they can be used in grouping terms.
        new_schema_length = len(op.scheme())
        implicit_group_by_cols = range(orig_schema_length, new_schema_length)

        ################################################
        # Compile away unbox expressions in where, emit clauses
        ################################################

        if where_clause:
            where_clause = multiway.rewrite_refs(where_clause, from_args, info)
            # Extract the type of there where clause to force type safety
            # to be checked
            where_clause.typeof(op.scheme(), None)
            op = raco.algebra.Select(condition=where_clause, input=op)

        emit_args = [(name, multiway.rewrite_refs(sexpr, from_args, info))
                     for (name, sexpr) in emit_args]

        statemods = multiway.rewrite_statemods(statemods, from_args, info)

        if any(raco.expression.expression_contains_aggregate(ex)
               for name, ex in emit_args):
            op = groupby.groupby(op, emit_args, implicit_group_by_cols,
                                 statemods)
        else:
            if statemods:
                return raco.algebra.StatefulApply(emit_args, statemods, op)
            if not (len(from_args) == 1 and len(emit_clause) == 1 and
                    isinstance(emit_clause[0],
                               (TableWildcardEmitArg, FullWildcardEmitArg))):
                op = raco.algebra.Apply(emit_args, op)

        if orderby_clause:
            if limit_clause is None:
                raise InvalidStatementException(
                    "An ORDER BY clause must be accompanied by a LIMIT clause")
            orderbyTuples = zip(*orderby_clause)
            op = raco.algebra.OrderBy(input=op,
                                      sort_columns=orderbyTuples[0],
                                      ascending=orderbyTuples[1])

        if limit_clause:
            if orderby_clause is None:
                raise InvalidStatementException(
                    "A LIMIT clause must be accompanied by an ORDER BY clause")
            op = raco.algebra.Limit(input=op, count=limit_clause)

        return op

    def distinct(self, expr):
        op = self.evaluate(expr)
        return raco.algebra.Distinct(input=op)

    def union(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("union", left, right)
        return raco.algebra.Union(left, right)

    def unionall(self, e1):
        return raco.algebra.UnionAll([self.evaluate(e) for e in e1])

    def countall(self, expr):
        op = self.evaluate(expr)
        grouping_list = []
        agg_list = [raco.expression.COUNTALL()]
        return raco.algebra.GroupBy(grouping_list, agg_list, op)

    def intersect(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("intersect", left, right)
        return raco.algebra.Intersection(left, right)

    def diff(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("diff", left, right)
        return raco.algebra.Difference(left, right)

    def limit(self, expr, count):
        op = self.evaluate(expr)
        return raco.algebra.Limit(input=op, count=count)

    def cross(self, left_target, right_target):
        left = self.evaluate(left_target)
        right = self.evaluate(right_target)

        return raco.algebra.CrossProduct(left, right)

    def join(self, left_target, right_target):
        """Convert parser.JoinTarget arguments into a Join operation"""

        left = self.evaluate(left_target.expr)
        right = self.evaluate(right_target.expr)

        assert len(left_target.columns) == len(right_target.columns)

        left_scheme = left.scheme()
        left_refs = [get_unnamed_ref(c, left_scheme, 0)
                     for c in left_target.columns]

        right_scheme = right.scheme()
        right_refs = [get_unnamed_ref(c, right_scheme, len(left_scheme))
                      for c in right_target.columns]

        join_conditions = [raco.expression.EQ(x, y) for x, y in
                           zip(left_refs, right_refs)]

        # Merge the join conditions into a big AND expression

        def andify(x, y):
            """Merge two scalar expressions with an AND"""
            return raco.expression.AND(x, y)

        condition = reduce(andify, join_conditions)
        return raco.algebra.Join(condition, left, right)


class StatementProcessor(object):

    """Evaluate a list of statements"""

    def __init__(self, catalog, use_dummy_schema=False):
        # Map from identifiers (aliases) to raco.algebra.Operation instances
        self.symbols = {}

        assert isinstance(catalog, raco.catalog.Catalog)
        self.catalog = catalog

        self.ep = ExpressionProcessor(self.symbols, catalog, use_dummy_schema)

        self.cfg = ControlFlowGraph()

    def evaluate(self, statements):
        """Evaluate a list of statements"""
        for statement in statements:
            # Switch on the first tuple entry
            method = getattr(self, statement[0].lower())
            method(*statement[1:])

    def __evaluate_expr(self, expr, _def):
        """Evaluate an expression; add a node to the control flow graph.

        :param expr: An expression to evaluate
        :type expr: Myrial AST tuple
        :param _def: The variable defined by the expression, or None for
                     non-statements
        :type _def: string
        """

        op = self.ep.evaluate(expr)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, _def, uses_set)
        return op

    def __do_assignment(self, _id, expr):
        """Process an assignment statement; add a node to the control flow
        graph.
        :param _id: The target variable name.
        :type _id: string
        :param expr: The relational expression to evaluate
        :type expr: A Myrial expression AST node tuple
        """

        child_op = self.ep.evaluate(expr)
        if _id in self.symbols:
            check_assignment_compatability(child_op, self.symbols[_id])

        op = raco.algebra.StoreTemp(_id, child_op)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, _id, uses_set)

        # Point future references of this symbol to a scan of the materialized
        # table. Note that this assumes there is no scoping in Myrial.
        self.symbols[_id] = raco.algebra.ScanTemp(_id, child_op.scheme())

    def assign(self, _id, expr):
        """Map a variable to the value of an expression."""
        self.__do_assignment(_id, expr)

    def idbassign(self, _id, agg, expr):
        """Map an IDB to the value of an expression."""
        self.__do_assignment(_id, expr)

    def store(self, _id, rel_key, how_distributed):
        assert isinstance(rel_key, relation_key.RelationKey)

        alias_expr = ("ALIAS", _id)
        child_op = self.ep.evaluate(alias_expr)

        if how_distributed == "BROADCAST":
            child_op = raco.algebra.Broadcast(child_op)
        elif how_distributed == "ROUND_ROBIN":
            child_op = raco.algebra.Shuffle(
                child_op, None, shuffle_type=Shuffle.ShuffleType.RoundRobin)
        # hash-partitioned
        elif how_distributed:
            scheme = child_op.scheme()
            col_list = [get_unnamed_ref(a, scheme) for a in how_distributed]
            child_op = raco.algebra.Shuffle(
                child_op, col_list, shuffle_type=Shuffle.ShuffleType.Hash)
        op = raco.algebra.Store(rel_key, child_op)

        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def sink(self, _id):
        alias_expr = ("ALIAS", _id)
        child_op = self.ep.evaluate(alias_expr)
        op = raco.algebra.Sink(child_op)

        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def dump(self, _id):
        alias_expr = ("ALIAS", _id)
        child_op = self.ep.evaluate(alias_expr)
        op = raco.algebra.Dump(child_op)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def dowhile(self, statement_list, termination_ex):
        first_op_id = self.cfg.next_op_id  # op ID of the top of the loop

        for _type, _id, expr in statement_list:
            if _type != 'ASSIGN':
                # TODO: Better error message
                raise InvalidStatementException('%s not allowed in do/while' %
                                                _type.lower())
            self.__do_assignment(_id, expr)

        last_op_id = self.cfg.next_op_id

        self.__evaluate_expr(termination_ex, None)

        # Add a control flow edge from the loop condition to the top of the
        # loop
        self.cfg.add_edge(last_op_id, first_op_id)

    def check_schema(self, expr):
        return True

    def get_idb_leaves(self, expr, idbs):
        ret = []
        op = expr[0].lower()
        args = expr[1:]
        if op in ["bagcomp"]:
            for _id, arg in args[0]:
                if arg:
                    ret += self.get_idb_leaves(arg, idbs)
                elif _id in idbs:
                    ret += [_id]
        elif op in ["select"]:
            for _id, arg in args[0].from_:
                if arg:
                    ret += self.get_idb_leaves(arg, idbs)
                elif _id in idbs:
                    ret += [_id]
        elif op in ["join", "union", "cross", "diff", "intersect"]:
            ret += self.get_idb_leaves(args[0], idbs)
            ret += self.get_idb_leaves(args[1], idbs)
        elif op in ["unionall"]:
            for child in args[0]:
                ret += self.get_idb_leaves(child, idbs)
        elif op in ["limit", "countall", "distinct"]:
            ret += self.get_idb_leaves(args[0], idbs)
        elif op in ["alias"]:
            if args[0] in idbs:
                ret += [args[0]]
        else:
            raise InvalidStatementException('%s not recognized' % op)
        return ret

    def separate_inputs(self, expr, idbs, is_init):
        op = expr[0].lower()
        if op in ["unionall"]:
            inputs = []
            for input in expr[1]:
                inputs += self.separate_inputs(input, idbs, is_init)
            return inputs
        else:
            edb_only = len(self.get_idb_leaves(expr, idbs)) == 0
            if (is_init and edb_only) or (not is_init and not edb_only):
                return [expr]
        return []

    def untilconvergence(self, statement_list, recursion_mode,
                         pull_order_policy):
        idbs = {}
        idx = 0
        for _type, _id, emits, expr in statement_list:
            if _id in self.symbols:
                raise InvalidStatementException('IDB %s is already used' % _id)
            idbcontroller = raco.algebra.IDBController(
                _id, idx,
                [None, None, raco.algebra.EmptyRelation(raco.scheme.Scheme())],
                emits, None, recursion_mode)
            idbs[_id] = idbcontroller
            self.symbols[_id] = raco.algebra.ScanIDB(_id, None, idbcontroller)
            idx = idx + 1

        for _type, _id, emits, expr in statement_list:
            initial_inputs = self.separate_inputs(expr, idbs, True)
            if len(initial_inputs) == 0:
                idbs[_id].children()[0] =\
                    raco.algebra.EmptyRelation(raco.scheme.Scheme())
            elif len(initial_inputs) == 1:
                idbs[_id].children()[0] = self.ep.evaluate(initial_inputs[0])
            else:
                idbs[_id].children()[0] =\
                    raco.algebra.UnionAll([self.ep.evaluate(expr)
                                           for expr in initial_inputs])

        done = False
        while (not done):
            done = True
            for _type, _id, emits, expr in statement_list:
                if idbs[_id].children()[1] is not None:
                    continue
                leaves = self.get_idb_leaves(expr, idbs)
                if any(idbs[leaf].scheme() is None for leaf in leaves):
                    done = False
                else:
                    iterative_inputs = self.separate_inputs(expr, idbs, False)
                    if len(iterative_inputs) == 0:
                        idbs[_id].children()[1] = raco.algebra.EmptyRelation(
                            raco.scheme.Scheme())
                    elif len(iterative_inputs) == 1:
                        idbs[_id].children()[1] = self.ep.evaluate(
                            iterative_inputs[0])
                    else:
                        idbs[_id].children()[1] = raco.algebra.UnionAll(
                            [self.ep.evaluate(expr)
                             for expr in iterative_inputs])

        op = raco.algebra.UntilConvergence(idbs.values(), pull_order_policy)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def get_logical_plan(self, **kwargs):
        """Return an operator representing the logical query plan."""
        return self.cfg.get_logical_plan(
            dead_code_elimination=kwargs.get('dead_code_elimination', True),
            apply_chaining=kwargs.get('apply_chaining', True))

    def __get_physical_plan_for__(self, target_phys_algebra, **kwargs):
        logical_plan = self.get_logical_plan(**kwargs)

        kwargs['target'] = target_phys_algebra
        return optimize(logical_plan, **kwargs)

    def get_physical_plan(self, **kwargs):
        """Return an operator representing the physical query plan."""
        target_phys_algebra = kwargs.get('target_alg')
        if target_phys_algebra is None:
            if kwargs.get('multiway_join', False):
                target_phys_algebra = MyriaHyperCubeAlgebra(self.catalog)
            else:
                target_phys_algebra = MyriaLeftDeepTreeAlgebra()

        return self.__get_physical_plan_for__(target_phys_algebra, **kwargs)

    def get_json(self, **kwargs):
        lp = self.get_logical_plan()
        pps = self.get_physical_plan(**kwargs)

        # TODO This is not correct. The first argument is the raw query string,
        # not the string representation of the logical plan
        return compile_to_json(str(lp), pps, pps, "myrial")

    @classmethod
    def get_json_from_physical_plan(cls, pp):
        pps = pp

        # TODO This is not correct. The first argument is the raw query string,
        # not the string representation of the logical plan
        return compile_to_json(
            "NOT_SOURCED_FROM_LOGICAL_RA", pps, pps, "myrial")



# coding=utf-8
""" Base class for Python lambda-oriented unit tests"""

import json
import unittest

import raco.fakedb
import raco.viz
import raco.myrial.interpreter as interpreter
from raco.backends.myria import compile_to_json
from raco.fake_data import FakeData
from raco import relation_key
from raco.replace_with_repr import replace_with_repr
from raco.algebra import Sequence, Scan, Store
from raco import compile
from raco.backends.logical import OptLogicalAlgebra
from raco.backends.myria import MyriaLeftDeepTreeAlgebra


class PythonTestCase(unittest.TestCase, FakeData):
    def setUp(self):
        self.db = raco.fakedb.FakeDatabase()
        self.processor = interpreter.StatementProcessor(self.db)
        self.db.ingest(self.emp_key, self.emp_table, self.emp_schema)
        self.input = relation_key.RelationKey.from_string(self.emp_key)
        self.schema = self.emp_schema
        self.scan = Scan(self.input, self.schema)
        self.output = relation_key.RelationKey.from_string('OUTPUT')

    def get_query(self, expression):
        return Sequence([Store(self.output, expression)])

    def get_plan(self, statements, logical=False):
        """Get the query plan"""
        self.logical = compile.optimize(statements, OptLogicalAlgebra())
        self.physical = compile.optimize(self.logical,
                                         MyriaLeftDeepTreeAlgebra())
        plan = self.logical if logical else self.physical

        # Verify that we can stringify
        assert str(plan)
        # Verify that we can convert to a dot
        raco.viz.get_dot(plan)
        # verify repr
        return replace_with_repr(plan)

    def execute_query(self, query, test_logical=False, skip_json=False):
        """Run a test query against the fake database"""
        plan = self.get_plan(query, logical=test_logical)

        if not test_logical and not skip_json:
            # Test that JSON compilation runs without error
            json_string = json.dumps(compile_to_json("some query",
                                                     "some logical plan",
                                                     plan,
                                                     "myrial"))
            assert json_string

        self.db.evaluate(plan)

        return self.db.get_table(self.output)

    def check_result(self, query, expected, test_logical=False,
                     skip_json=False, scheme=None):
        """Execute a test query with an expected output"""
        actual = self.execute_query(query, test_logical, skip_json)
        self.assertEquals(actual, expected)

        if scheme:
            self.assertEquals(self.db.get_scheme(self.output), scheme)

    def check_scheme(self, query, scheme):
        """Execute a test query with an expected output schema."""
        self.execute_query(query)
        self.assertEquals(self.db.get_scheme(self.output), scheme)



# coding=utf-8

""" Tests for decompiling lambdas """

import unittest

from raco.python.util import decompile


class TestDecompileFunctions(unittest.TestCase):
    def test_simple(self):
        def f():
            def g(): return 0
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_space_before_colon(self):
        def f() :
            def g(): return 0
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_multiple_lines(self):
        def f() :
            def g():

                return 0
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_line_continuation(self):
        def f() :
            def g():
                return \
                    0
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_commants(self):
        def f() :
            def g(): # foo
                return \
                    0 # bar
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_variable(self):
        def f():
            def g(): return 0
            return g
        h = f()
        s = decompile.get_source(h)
        exec s
        self.assertEquals(g(), 0)

    def test_function_with_lambda(self):
        def f():
            def g(x=lambda: 0): return x()
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), 0)

    def test_embedded_lambda_token(self):
        def f():
            def g(x="lambda: 0"): return "lambda: 1"
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(), "lambda: 1")

    def test_parameters(self):
        def f():
            def g(x, y, z=1): return x + y + z
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(3, 4, 5), 12)
        self.assertEquals(g(3, 4), 8)

    def test_args_kwargs(self):
        def f():
            def g(*args, **kwargs): return args[0] + kwargs['foo']
            return g
        s = decompile.get_source(f())
        exec s
        self.assertEquals(g(3, foo=4), 7)

    def _member(self): return 1

    def test_member(self):
        s = decompile.get_source(self._member)
        exec s
        self.assertEquals(_member(self), 1)

    @classmethod
    def _classmethod(self): return 1

    def test_classmethod(self):
        s = decompile.get_source(self._classmethod)
        exec s
        self.assertEquals(_classmethod(self), 1)

    @staticmethod
    def _staticmethod(x): return x + 1

    def test_staticmethod(self):
        s = decompile.get_source(self._staticmethod)
        exec s
        self.assertEquals(_staticmethod(3), 4)

    def test_unpacking(self):
        def fff((x, y)): return x + y
        s = decompile.get_source(fff)
        exec s.replace('fff', 'ggg')
        self.assertEquals(ggg((1, 2)), 3)




import logging

import raco.rules

from raco import algebra
from raco.scheme import Scheme
from raco.backends import Language, Algebra
from raco.expression import UnnamedAttributeRef, \
    NamedAttributeRef, \
    AttributeRef

LOGGER = logging.getLogger(__name__)

# for gensym
counter = 0


class SPARQLLanguage(Language):
    EQ = "="

    @classmethod
    def body(cls, expr):
        return expr.compile()

    @classmethod
    def compile_attribute(cls, expr):
        if isinstance(expr, NamedAttributeRef):
            var = expr
        if isinstance(expr, UnnamedAttributeRef):
            var = cls.currentop.scheme().getName(expr.position)
        return "?%s" % var

    @classmethod
    def compile_stringliteral(cls, value):
        return str(value).replace('"', '')


class SPARQLOperator(object):
    language = SPARQLLanguage


class SPARQLUnionAll(algebra.UnionAll, SPARQLOperator):

    def compile(self):
        return """
      {
      %s
      }
      UNION
      {
      %s
      }
      """ % (self.left.compile(), self.right.compile())


class SPARQLScan(algebra.Scan, SPARQLOperator):

    def renameattrs(self):
        """Make attribute names globally unique so they
        can be used as SPARQL variables"""
        global counter
        c = counter
        counter = counter + 1
        self._scheme = Scheme([("%s%s" % (n, c), typ)
                               for n, typ in self._scheme])

    def __init__(self, *args):
        # Python 3.0 cleans this crap up
        super(self.__class__, self).__init__(*args)
        if self._scheme:
            self.renameattrs()

    def copy(self, other):
        # Python 3.0 cleans this crap up
        super(self.__class__, self).copy(other)
        self.renameattrs()

    def compile(self):
        names = self.scheme().get_names()
        triplepattern = " ".join(["?%s" % n for n in names])
        return "%s ." % (triplepattern)


class SPARQLSelect(algebra.Select, SPARQLOperator):

    def compile(self):
        # This is pretty bad: passing information to compile_expression
        # by way of class instance variable
        # The right thing to do seems to be to arrange for every
        # AttributeRef to hold a pointer to the operator
        # whose result is being referenced.  This way, if
        # you have an AttributeRef in hand,
        # you know you can do something useful with it.
        self.language.currentop = self
        filterexpr = self.language.compile_expression(self.condition)
        return """%s
    FILTER (%s) .""" % (self.input.compile(), filterexpr)


class SPARQLStore(algebra.Store, SPARQLOperator):

    def compile(self):
        return """%s""" % (self.input.compile())


class SPARQLStoreTemp(SPARQLStore):
    pass


class SPARQLJoin(algebra.Join, SPARQLOperator):

    def compile(self):
        leftc = self.left.compile()
        rightc = self.right.compile()
        self.language.currentop = self
        joincond = self.language.compile_expression(self.condition)
        return """
        {
        %s
        }
        {
        %s
        }
        FILTER (%s)
        """ % (self.left.compile(), self.right.compile(), joincond)


class SPARQLApply(algebra.Apply, SPARQLOperator):

    """Represents a simple apply operator"""

    def compile(self):
        child_scheme = self.input.scheme()
        self.language.currentop = self.input

        def formatemitter(name, exp):
            e = self.language.compile_expression(exp)
            if isinstance(exp, AttributeRef):
                return "%s" % e
            else:
                return "%s as ?%s" % (e, name)
        emitters = ", ".join([formatemitter(name, exp)
                              for (name, exp) in self.emitters])
        return """
  SELECT %s
  {
   %s
  }
  """ % (emitters, self.input.compile())


class SPARQLAlgebra(Algebra):

    def opt_rules(self, **kwargs):
        rules = [
            raco.rules.CrossProduct2Join(),
            raco.rules.SplitSelects(),
            raco.rules.PushSelects(),
            raco.rules.DeDupBroadcastInputs(),
            raco.rules.OneToOne(algebra.Scan, SPARQLScan),
            raco.rules.OneToOne(algebra.Store, SPARQLStore),
            raco.rules.OneToOne(algebra.Select, SPARQLSelect),
            raco.rules.OneToOne(algebra.Apply, SPARQLApply),
            raco.rules.OneToOne(algebra.Join, SPARQLJoin),
        ]

        return rules



# coding=utf-8
""" Tests for projection expressions """

from collections import Counter
from raco.python import convert
from raco.algebra import Apply
from python_test import PythonTestCase


class TestProjection(PythonTestCase):
    def _execute_projection(self, query, expected):
        projection = convert(query, [self.schema])
        self.assertIsNotNone(projection)

        expression = Apply([('out', projection)], self.scan)
        plan = self.get_query(expression)
        return self.check_result(plan, expected)

    def test_name(self):
        self._execute_projection("""lambda t: t.name""",
                                 Counter([(t[2],) for t in self.emp_table]))

    def test_expression(self):
        self._execute_projection("""lambda t: t.id + 1""",
                                 Counter([(t[0] + 1,) for t
                                          in self.emp_table]))

    def test_index(self):
        self._execute_projection("""lambda t: t[2]""",
                                 Counter([(t[2],) for t in self.emp_table]))



# coding=utf-8

""" Tests for decompiling lambdas """

import unittest

from raco.python.exceptions import PythonConvertException
from raco.python.util import decompile


class TestDecompileLambdas(unittest.TestCase):
    def test_simple(self):
        s = decompile.get_source(lambda: 0)
        self.assertEquals(eval(s)(), 0)

    def test_no_space(self):
        s = decompile.get_source(lambda:0)
        self.assertEquals(eval(s)(), 0)

    def test_space_after_colon(self):
        s = decompile.get_source(lambda :0)
        self.assertEquals(eval(s)(), 0)

    def test_variable(self):
        f = lambda: 0
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(), f())

    def test_newline(self):
        f = \
            lambda: 0
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(), f())

    def test_newline2(self):
        f = lambda: \
            0
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(), f())

    def test_tuple(self):
        t = (lambda: 0), 5
        s = decompile.get_source(t[0])
        self.assertEquals(eval(s)(), t[0]())

    def test_tuple2(self):
        f = lambda :  (0, 5)
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(), f())

    def test_multiple_lambdas(self):
        t = lambda: 1, lambda: 2, lambda: 3
        for f in t:
            s = decompile.get_source(f)
            self.assertEquals(eval(s)(), f())

    def test_embedded_lambda_token(self):
        f = lambda: "lambda: 0"
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(), f())

    def test_parameters(self):
        f = lambda x: x
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(5), f(5))

    def test_multiple_parameters(self):
        f = lambda x, y: x + y
        s = decompile.get_source(f)
        self.assertEquals(eval(s)(5, 6), f(5, 6))

    def test_args_kwargs(self):
        f = lambda *args, **kwargs: args[0] + kwargs['foo']
        s = decompile.get_source(f)
        self.assertEqual(eval(s)(5, foo=6), f(5, foo=6))

    def test_unpacking(self):
        """ Unpacking is not currently supported """
        f = lambda (x, y): x + y
        self.assertRaises(PythonConvertException,
                          lambda: decompile.get_source(f))



# coding=utf-8
""" Utilities for converting compiled functions to Python source """

import ast
import inspect
import re
from types import CodeType

from raco.python.exceptions import PythonConvertException

lambda_header = 'lambda:'        # Prefix associated with a lambda
module_name = '<module>'         # Name of a non-lambda code unit
min_length = len(lambda_header)  # shortest lambda expression possible
lambda_expression = r'(?:^|\W)(?=(lambda[^:]*:.*))'
function_expression = r'(?:^|\W)(?=(def\W[^:]*:.*))'


def get_source(f):
    """ Return the source of a lambda function """
    try:
        lines, _ = inspect.getsourcelines(f)
    except (IOError, TypeError) as e:
        raise PythonConvertException(str(e))

    source = ' '.join(line.strip(' ') for line in lines).strip('\n\r ')
    candidates = _get_candidates(source)

    try:
        return next(source for candidate in candidates
                    for source in [_get_source(f, candidate)]
                    if source)
    except StopIteration:
        raise PythonConvertException('Unable to convert function into source')


def _get_source(f, source):
    """
    Convert the given source into an AST
    :param source: Source code of a function or lambda expression
    :return: An AST representing the source
    """
    for source in _reverse_walk_source(source):
        try:
            tree = ast.parse(source)
            node = _get_definition_node(tree)
            return _verify_ast(f, node, source)
        except SyntaxError:
            pass

    return None


def _verify_ast(f, tree, source):
    """
    Verify that a given tree matches the original function
    AST nodes maintain their starting offsets in the input source,
    so we have to work backwards to determine the actual end.

    :param tree: an AST representation of a function or lambda
    :return: True when the AST matches the function or lambda; false otherwise
    """

    for source, body in zip(
            _reverse_walk_source(source[tree.col_offset:], min_length),
            _reverse_walk_source(_get_body_source(tree, source))):
        try:
            recompiled_f = _compile(source, body)
            if _verify_bytecode(source, f, recompiled_f):
                return source
        except SyntaxError:
            pass

    return None


def _compile(source, body):
    try:
        return compile(body, '', 'eval')
    except SyntaxError:
        module = compile(source, '', 'exec')
        function = filter(None, module.co_consts)[-1]  # Best guess
        if module.co_name != module_name:
            raise PythonConvertException('Function must compile to a module')
        elif not isinstance(function, CodeType):
            raise PythonConvertException('Module must contain function')
        return function


def _get_body_source(tree, source):
    first_statement = tree.body[0] if isinstance(tree.body, list) \
        else tree.body
    line_offset = sum(len(line) for line
                      in source.split('\n')[:first_statement.lineno - 1])
    return source[line_offset + first_statement.col_offset - 1:].strip(': ')


def _verify_bytecode(source, f, recompiled_f, require_callable=True):
    try:
        return ((not require_callable or
                 callable(eval(source))) and
                filter(None, f.func_code.co_consts) ==
                filter(None, recompiled_f.co_consts) and
                len(f.__code__.co_code) == len(recompiled_f.co_code))
    except NameError:
        return False
    except SyntaxError as e:
        if not source.startswith(lambda_header):
            return _verify_bytecode(source, f, recompiled_f, False) \
                if isinstance(recompiled_f, CodeType) \
                else False
        else:
            raise e


def _reverse_walk_source(source, min_length=0):
    while len(source) > min_length:
        yield source
        source = source[:-1].strip()


def _get_candidates(source):
    """
    Get all possible lambda expressions in a source string,
    where the results may contain incorrect trailing characters
    (we'll remove these later)

    E.g.: lambda x: 'lambda:0' -> ["lambda x: 'lambda:0'", "lambda:0'"]
    """
    lambdas = re.findall(lambda_expression, source, re.MULTILINE | re.DOTALL)
    functions = re.findall(function_expression, source,
                           re.MULTILINE | re.DOTALL)
    return lambdas + functions


def _get_definition_node(tree):
    return next((node for node in ast.walk(tree) if
                 isinstance(node, ast.Lambda) or
                 isinstance(node, ast.FunctionDef)))



"""
A RACO language to compile expressions to SQL.
"""

from sqlalchemy import (Column, Table, MetaData, Integer, String,
                        Float, Boolean, LargeBinary, DateTime, select, func,
                        literal, case)

import raco.algebra as algebra
from raco.catalog import Catalog
import raco.expression as expression
import raco.scheme as scheme
import raco.types as types
from raco.representation import RepresentationProperties
import abc


type_to_raco = {Integer: types.LONG_TYPE,
                String: types.STRING_TYPE,
                Float: types.FLOAT_TYPE,
                Boolean: types.BOOLEAN_TYPE,
                LargeBinary: types.BLOB_TYPE,
                DateTime: types.DATETIME_TYPE}


raco_to_type = {types.LONG_TYPE: Integer,
                types.INT_TYPE: Integer,
                types.STRING_TYPE: String,
                types.FLOAT_TYPE: Float,
                types.DOUBLE_TYPE: Float,
                types.BOOLEAN_TYPE: Boolean,
                types.BLOB_TYPE: LargeBinary,
                types.DATETIME_TYPE: DateTime}


class SQLFunctionProvider(object):
    """Interface for translating function names. For Raco functions
    not understood by SQLAlchemy, like stdev, we cannot rely
    on SQLAlchemy's compiler to translate function
    names to the given dialect.
    For functions not understood by SQLAlchemy, the SQLAlchemy compiler
    just emits them verbatim."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def convert_unary_expr(self, expr, input):
        pass


class _DefaultSQLFunctionProvider(SQLFunctionProvider):
    def convert_unary_expr(self, expr, input):
        # just use the function name without complaining
        fname = expr.__class__.__name__.lower()
        return getattr(func, fname)(input)


class PostgresSQLFunctionProvider(SQLFunctionProvider):
    def convert_unary_expr(self, expr, input):
        fname = expr.__class__.__name__.lower()

        # replacements
        if fname == "stdev":
            return func.stddev_samp(input)

        # Warning: may create some functions not available in Postgres
        return getattr(func, fname)(input)


class SQLCatalog(Catalog):
    def __init__(self, engine=None, push_grouping=False,
                 provider=_DefaultSQLFunctionProvider()):
        self.engine = engine
        self.push_grouping = push_grouping
        self.provider = provider
        self.metadata = MetaData()

    @staticmethod
    def get_num_servers():
        """ Return number of servers in myria deployment """
        return 1

    def num_tuples(self, rel_key):
        """ Return number of tuples of rel_key """
        table = self.metadata.tables[str(rel_key)]
        return self.engine.execute(table.count()).scalar()

    def partitioning(self, rel_key):
        return RepresentationProperties()

    def get_scheme(self, rel_key):
        table = self.metadata.tables[str(rel_key)]
        return scheme.Scheme((c.name, type_to_raco[type(c.type)])
                             for c in table.columns)

    def add_table(self, name, schema):
        columns = [Column(n, raco_to_type[t](), nullable=False)
                   for n, t in schema.attributes]
        # Adds the table to the metadata
        Table(name, self.metadata, *columns)

    def add_tuples(self, name, schema, tuples=None):
        table = self.metadata.tables[name]
        table.create(self.engine)
        if tuples:
            tuples = [{n: v for n, v in zip(schema.get_names(), tup)}
                      for tup in tuples]
            self.engine.execute(table.insert(), tuples)

    def _convert_expr(self, cols, expr, input_scheme):
        if isinstance(expr, expression.AttributeRef):
            return self._convert_attribute_ref(cols, expr, input_scheme)
        if isinstance(expr, expression.ZeroaryOperator):
            return self._convert_zeroary_expr(cols, expr, input_scheme)
        if isinstance(expr, expression.UnaryOperator):
            return self._convert_unary_expr(cols, expr, input_scheme)
        if isinstance(expr, expression.BinaryOperator):
            return self._convert_binary_expr(cols, expr, input_scheme)

        if isinstance(expr, expression.Case):
            conv = lambda e: self._convert_expr(cols, e, input_scheme)
            conditions = [(conv(when), conv(then))
                          for when, then in expr.when_tuples]
            else_result = conv(expr.else_expr)
            return case(conditions, else_=else_result)

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_attribute_ref(self, cols, expr, input_scheme):
        if isinstance(expr, expression.NamedAttributeRef):
            expr = expression.toUnnamed(expr, input_scheme)

        if isinstance(expr, expression.UnnamedAttributeRef):
            # Not an elif since the first may actually turn into a UARef
            return cols[expr.position]

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_zeroary_expr(self, cols, expr, input_scheme):
        if isinstance(expr, expression.COUNTALL):
            return func.count(cols[0])
        if isinstance(expr, expression.Literal):
            return literal(expr.value,
                           raco_to_type[expr.typeof(input_scheme, None)])
        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_unary_expr(self, cols, expr, input_scheme):
        input = self._convert_expr(cols, expr.input, input_scheme)

        c = self.provider.convert_unary_expr(expr, input)
        if c is not None:
            return c

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_binary_expr(self, cols, expr, input_scheme):
        left = self._convert_expr(cols, expr.left, input_scheme)
        right = self._convert_expr(cols, expr.right, input_scheme)

        if isinstance(expr, expression.AND):
            return left & right
        if isinstance(expr, expression.OR):
            return left | right
        if isinstance(expr, expression.EQ):
            return left == right
        if isinstance(expr, expression.NEQ):
            return left != right
        if isinstance(expr, expression.LT):
            return left < right
        if isinstance(expr, expression.LTEQ):
            return left <= right
        if isinstance(expr, expression.GT):
            return left > right
        if isinstance(expr, expression.GTEQ):
            return left >= right

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _get_zeroary_sql(self, plan):
        if isinstance(plan, algebra.Scan):
            if str(plan.relation_key) not in self.metadata.tables:
                self.add_table(str(plan.relation_key), plan.scheme())
            return self.metadata.tables[str(plan.relation_key)].select()
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_unary_sql(self, plan):
        input = self.get_sql(plan.input).alias("input")
        input_sch = plan.input.scheme()
        cols = list(input.c)

        if isinstance(plan, algebra.Select):
            cond = self._convert_expr(cols, plan.condition, input_sch)
            return input.select(cond)

        elif isinstance(plan, algebra.Apply):
            clause = [self._convert_expr(cols, e, input_sch).label(name)
                      for (name, e) in plan.emitters]
            return select(clause, from_obj=input)

        elif isinstance(plan, algebra.GroupBy):
            if (not self.push_grouping) and len(plan.grouping_list) > 0:
                raise NotImplementedError(
                    """convert aggregate with grouping to sql
                    -- Myria is faster. If you want to push group by into
                    SQL use the flag push_sql_grouping""")
            a = [self._convert_expr(cols, e, input_sch)
                 for e in plan.aggregate_list]
            g = [self._convert_expr(cols, e, input_sch)
                 for e in plan.grouping_list]
            sel = select(g + a)
            if not plan.grouping_list:
                return sel
            return sel.group_by(*g)

        elif isinstance(plan, algebra.Distinct):
            return select(['*'], from_obj=input, distinct=True)

        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_binary_sql(self, plan):
        # Use aliases to resolve duplicate names
        left = self.get_sql(plan.left).alias("left")
        right = self.get_sql(plan.right).alias("right")
        all_cols = left.c + right.c
        all_sch = plan.left.scheme() + plan.right.scheme()
        assert len(all_cols) == len(all_sch)

        if isinstance(plan, algebra.CrossProduct):
            out_cols = zip(all_cols, all_sch.get_names())
            return select([col.label(name) for col, name in out_cols])

        elif isinstance(plan, algebra.ProjectingJoin):
            cond = self._convert_expr(all_cols, plan.condition, all_sch)

            clause = [self._convert_expr(all_cols, e, all_sch)
                      for e in plan.output_columns]

            return select(clause, from_obj=left.join(right, cond))

        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_nary_sql(self, plan):
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def get_sql(self, plan):
        if isinstance(plan, algebra.ZeroaryOperator):
            return self._get_zeroary_sql(plan)
        elif isinstance(plan, algebra.UnaryOperator):
            return self._get_unary_sql(plan)
        elif isinstance(plan, algebra.BinaryOperator):
            return self._get_binary_sql(plan)
        elif isinstance(plan, algebra.NaryOperator):
            return self._get_nary_sql(plan)
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def evaluate(self, plan):
        statement = self.get_sql(plan)
        print statement
        return (tuple(t) for t in self.engine.execute(statement))



import base64
import ConfigParser
import json
import csv
from time import sleep
import logging
from urlparse import urlparse, ParseResult

from raco.backends.logical import OptLogicalAlgebra
from raco.backends.myria import MyriaHyperCubeAlgebra, MyriaLeftDeepTreeAlgebra, \
    compile_to_json
from raco.backends.myria.catalog import MyriaCatalog
from raco.myrial import interpreter
from raco.myrial.parser import Parser

from .errors import MyriaError

import requests

from raco import compile, RACompiler

__all__ = ['MyriaConnection']

# String constants used in forming requests
JSON = 'application/json'
CSV = 'text/plain'
GET = 'GET'
PUT = 'PUT'
POST = 'POST'

# Enable or configure logging
logging.basicConfig(level=logging.WARN)


class FunctionTypes(object):
    POSTGRES = 0
    PYTHON = 1


class MyriaConnection(object):
    """Contains a connection the Myria REST server."""

    _DEFAULT_HEADERS = {
        'Accept': JSON,
        'Content-Type': JSON
    }

    @staticmethod
    def _parse_deployment(deployment):
        "Extract the REST server hostname and port from a deployment.cfg file"
        if deployment is None:
            return None
        config = ConfigParser.RawConfigParser(allow_no_value=True)
        config.readfp(deployment)
        master = config.get('master', '0')
        hostname = master[:master.index(':')]
        port = int(config.get('deployment', 'rest_port'))
        return (hostname, port)

    def __init__(self,
                 deployment=None,
                 hostname=None,
                 port=None,
                 ssl=False,
                 rest_url=None,
                 execution_url=None,
                 timeout=None):
        """Initializes a connection to the Myria REST server.
           (And optionally a Myria program execution URI.)

        Args:
            deployment: An open file (or other buffer) containing a
                deployment.cfg file in Myria's format. This file will be parsed
                to determine the REST server hostname and port.
            hostname: The hostname of the REST server. May be overwritten if
                deployment is provided.
            port: The port of the REST server. May be overwritten if deployment
                is provided.
            timeout: The timeout for the connection to myria.

            rest_url: a URL pointing to a Myria REST endpoint
            execution_url: a URL pointing to a Myria webserver for program
                execution
        """
        # Parse the deployment file and, if present, override the hostname and
        # port with any provided values from deployment.
        rest_config = self._parse_deployment(deployment)
        if rest_config is not None:
            hostname = hostname or rest_config[0]
            port = port or rest_config[1]
        if rest_url is not None:
            url = urlparse(rest_url)
            hostname = url.hostname
            ssl = url.scheme == "https"
            port = url.port or (443 if ssl else 80)

        if ssl:
            uri_scheme = "https"
        else:
            uri_scheme = "http"

        if execution_url is None:
            execution_url = ParseResult(scheme=uri_scheme,
                                        netloc=hostname,
                                        path="", params="",
                                        query="", fragment="").geturl()

        self._url_start = '{}://{}:{}'.format(uri_scheme, hostname, port)
        self._session = requests.Session()
        self._session.headers.update(self._DEFAULT_HEADERS)
        self.execution_url = execution_url
        self._udfs = None

    def _finish_async_request(self, method, url, body=None, accept=JSON):
        headers = {
            'Accept': accept
        }
        try:
            while True:
                if '://' not in url:
                    url = self._url_start + url
                logging.info("Finish async request to {}. Headers: {}".format(
                    url, headers))
                r = self._session.request(method, url, headers=headers,
                                          data=body)
                if r.status_code in [200, 201]:
                    if accept == JSON:
                        return r.json()
                    else:
                        return r.text
                elif r.status_code in [202]:
                    # Get the new URL to poll, etc.
                    url = r.headers['Location']
                    method = GET
                    body = None
                    # Read and ignore the body
                    # response.read()
                    # Sleep 100 ms before re-issuing the request
                    sleep(0.1)
                else:
                    raise MyriaError('Error %d: %s'
                                     % (r.status_code, r.text))
        except Exception as e:
            if isinstance(e, MyriaError):
                raise
            raise MyriaError(e)

    def _make_request(self, method, url, body=None, params=None,
                      accept=JSON, get_request=False):
        headers = {
            'Accept': accept
        }
        if '://' not in url:
            url = self._url_start + url
        r = self._session.request(method, url, headers=headers,
                                  data=body, params=params, stream=True)
        logging.info("Make myria request to {}. Headers: {}".format(
                     r.url, headers))
        if r.status_code in [200, 201, 202]:
            if get_request:
                return r
            if accept == JSON:
                try:
                    return r.json()
                except ValueError, e:
                    raise MyriaError(
                        'Error %d: %s' % (r.status_code, r.text))
            else:
                return r.iter_lines()
        else:
            raise MyriaError('Error %d: %s'
                             % (r.status_code, r.text))

    def _wrap_get(self, selector, params=None, status=None, accepted=None):
        if status is None:
            status = [200]
        if accepted is None:
            accepted = []

        if '://' not in selector:
            selector = self._url_start + selector
        r = self._session.get(selector, params=params)
        if r.status_code in status:
            return r.json()
        elif r.status_code in accepted:
            return self._wrap_get(selector, params=params, status=status,
                                  accepted=accepted)
        else:
            raise MyriaError(r)

    def _wrap_post(self, selector, data=None, params=None, status=None,
                   accepted=None):
        if status is None:
            status = [201, 202]
            if accepted is None:
                accepted = [202]
        else:
            if accepted is None:
                accepted = []

        if '://' not in selector:
            selector = self._url_start + selector
        r = self._session.post(selector, data=data, params=params)
        if r.status_code in status:
            if r.headers['Location']:
                return self._wrap_get(r.headers['Location'], status=status,
                                      accepted=accepted)
            return r.json()
        else:
            raise MyriaError(r)

    def workers(self):
        """Return a dictionary of the workers"""
        return self._wrap_get('/workers')

    def workers_alive(self):
        """Return a list of the workers that are alive"""
        return self._wrap_get('/workers/alive')

    def worker(self, worker_id):
        """Return information about the specified worker"""
        return self._wrap_get('/workers/worker-{}'.format(worker_id))

    def datasets(self):
        """Return a list of the datasets that exist"""
        return self._wrap_get('/dataset')

    def dataset(self, relation_key):
        """Return information about the specified relation"""
        return self._wrap_get('/dataset/user-{}/program-{}/relation-{}'.format(
            relation_key['userName'],
            relation_key['programName'],
            relation_key['relationName']))

    def download_dataset(self, relation_key):
        """Download the data in the dataset as json"""
        return self._wrap_get('/dataset/user-{}/program-{}/relation-{}/data'
                              .format(relation_key['userName'],
                                      relation_key['programName'],
                                      relation_key['relationName']),
                              params={'format': 'json'})

    @staticmethod
    def _ensure_schema(schema):
        return {'columnTypes': schema['columnTypes'],
                'columnNames': schema['columnNames']}

    @staticmethod
    def _ensure_relation_key(relation_key):
        return {'userName': relation_key['userName'],
                'programName': relation_key['programName'],
                'relationName': relation_key['relationName']}

    def create_empty(self, relation_key, schema):
        return self.upload_source(relation_key, schema, {'dataType': 'Empty'})

    def upload_fp(self, relation_key, schema, fp):
        """Upload the data in the supplied fp to the specified user and
        relation.

        Args:
            relation_key: A dictionary containing the destination relation key.
            schema: A dictionary containing the schema,
            fp: A file pointer containing the data to be uploaded.
        """

        data = base64.b64encode(fp.read())
        source = {'dataType': 'Bytes',
                  'bytes': data}
        return self.upload_source(relation_key, schema, source)

    def upload_source(self, relation_key, schema, source):
        body = {'relationKey': self._ensure_relation_key(relation_key),
                'schema': self._ensure_schema(schema),
                'source': source}

        return self._make_request(POST, '/dataset', json.dumps(body))

    def execute_program(self, program, language="MyriaL", server=None):
        """Execute the program in the specified language on Myria, polling
        its status until the query is finished. Returns the query status
        struct.

        Args:
            program: a Myria program as a string.
            language: the language in which the program is written
                      (default: MyriaL).
        """

        body = {"query": program, "language": language}
        r = requests.post((server or self.execution_url) + '/execute',
                          data=body)
        if r.status_code != 201:
            raise MyriaError(r)

        query_uri = r.json()['url']
        while True:
            r = requests.get(query_uri)
            if r.status_code == 200:
                return r.json()
            elif r.status_code == 202:
                # Sleep 100 ms before re-checking the status
                sleep(0.1)
                continue
            raise MyriaError(r)

    def compile_program(self, program, language="MyriaL", **kwargs):
        """Get a compiled plan for a given program.

        Args:
            program: a Myria program as a string.
            language: the language in which the program is written
                      (default: MyriaL).
        """
        logical = self._get_plan(program, language, 'logical', **kwargs)
        physical = self._get_plan(program, language, 'physical', **kwargs)
        compiled = compile_to_json(program, logical, physical, language)
        compiled['profilingMode'] = ["QUERY", "RESOURCE"] \
            if kwargs.get('profile', False) else []
        return compiled

    def submit_query(self, query):
        """Submit the query to Myria, and return the status including the URL
        to be polled.

        Args:
            query: a Myria physical plan as a Python object.
        """

        body = json.dumps(query)
        return self._wrap_post('/query', data=body)

    def execute_query(self, query):
        """Submit the query to Myria, and poll its status until it finishes.

        Args:
            query: a Myria physical plan as a Python object.
        """

        body = json.dumps(query)
        return self._finish_async_request(POST, '/query', body)

    def validate_query(self, query):
        """Submit the query to Myria for validation only.

        Args:
            query: a Myria physical plan as a Python object.
        """

        body = json.dumps(query)
        return self._make_request(POST, '/query/validate', body)

    def get_query_status(self, query_id):
        """Get the status of a submitted query.

        Args:
            query_id: the id of a submitted query
        """

        resource_path = '/query/query-%d' % int(query_id)
        return self._make_request(GET, resource_path)

    def get_query_plan(self, query_id, subquery_id):
        """Get the saved execution plan for a submitted query.

        Args:
            query_id: the id of a submitted query
            subquery_id: the subquery id within the specified query
        """

        resource_path = '/query/query-{q}/subquery-{s}'.format(
            q=int(query_id), s=int(subquery_id))
        return self._make_request(GET, resource_path)

    def get_fragment_ids(self, query_id, worker_id):
        """Get the number of fragments in a query plan.

        Args:
            query_id: the id of a submitted query
            worker_id: the id of a worker
        """
        status = self.get_query_status(query_id)
        if 'fragments' in status['plan']:
            fids = []
            for fragment in status['plan']['fragments']:
                if int(worker_id) in map(int, fragment['workers']):
                    fids.append(fragment['fragmentIndex'])
            return fids
        else:
            return []

    def get_sent_logs(self, query_id, fragment_id=None):
        """Get the logs for where data was sent.

        Args:
            query_id: the id of a submitted query
            fragment_id: the id of a fragment
        """
        resource_path = '/logs/sent?queryId=%d' % int(query_id)
        if fragment_id is not None:
            resource_path += '&fragmentId=%d' % int(fragment_id)
        response = self._make_request(GET, resource_path, accept=CSV)
        return csv.reader(response)

    def get_profiling_log(self, query_id, fragment_id=None):
        """Get the logs for operators.

        Args:
            query_id: the id of a submitted query
            fragment_id: the id of a fragment
        """
        resource_path = '/logs/profiling?queryId=%d' % int(query_id)
        if fragment_id is not None:
            resource_path += '&fragmentId=%d' % int(fragment_id)
        response = self._make_request(GET, resource_path, accept=CSV)
        return csv.reader(response)

    def get_profiling_log_roots(self, query_id, fragment_id):
        """Get the logs for root operators.

        Args:
            query_id: the id of a submitted query
            fragment_id: the id of a fragment
        """
        resource_path = '/logs/profilingroots?queryId=%d&fragmentId=%d' % (int(
            query_id), int(fragment_id))
        response = self._make_request(GET, resource_path, accept=CSV)
        return csv.reader(response)

    def queries(self, limit=None, max_id=None, min_id=None, q=None):
        """Get count and information about all submitted queries.

        Args:
            limit: the maximum number of query status results to return.
            max_id: the maximum query ID to return.
            min_id: the minimum query ID to return. Ignored if max_id is
                    present.
            q: a text search for the raw query string.
        """
        params = {}
        if limit is not None:
            params['limit'] = limit
        if max_id is not None:
            params['max'] = max_id
        if min_id is not None:
            params['min'] = min_id
        if q is not None:
            params['q'] = q

        resource_path = '/query'
        r = self._make_request(GET, resource_path, params=params,
                               get_request=True)
        return r.json()

    def upload_file(self, relation_key, schema, data, overwrite=None,
                    delimiter=None, binary=None, is_little_endian=None):
        """Upload a file in a streaming manner to Myria.

        Args:
            relation_key: relation to be created.
            schema: schema of the relation.
            data: the bytes to be uploaded.
            overwrite: optional boolean indicating that an existing relation
                should be overwritten. Myria default is False.
            delimiter: optional character which delimits a CSV file. Only valid
                if binary is False. Myria default is ','.
            binary: optional boolean indicating that the data is encoded as
                a packed binary. Myria default is False.
            is_little_endian: optional boolean indicating that the binary data
                is in little-Endian. Myria default is False.
        """

        from requests_toolbelt import MultipartEncoder

        fields = [('relationKey', relation_key), ('schema', schema),
                  ('overwrite', overwrite), ('delimiter', delimiter),
                  ('binary', binary), ('isLittleEndian', is_little_endian)]
        fields = [(name, (name, json.dumps(value), 'application/json'))
                  for (name, value) in fields]
        # data must be last
        if binary:
            data_type = 'application/octet-stream'
        else:
            data_type = 'text/plain'
        fields.append(('data', ('data', data, data_type)))

        m = MultipartEncoder(fields=fields)
        r = self._session.post(self._url_start + '/dataset', data=m,
                               headers={'Content-Type': m.content_type})
        if r.status_code not in (200, 201):
            raise MyriaError('Error %d: %s'
                             % (r.status_code, r.text))
        return r.json()

    def get_function(self, name):
        """ Get user defined function metadata """
        return self._wrap_get('/function/{}'.format(name))

    def create_function(self, d, overwrite_if_exists=False):
        """Register a User Defined Function with Myria """
        result = self._make_request(POST, '/function', json.dumps(d))
        Parser.add_python_udf(d.pop('name'), d.pop('outputType'),
                              overwrite_if_exists=overwrite_if_exists, **d)
        return result

    def get_functions(self):
        """ List all the user defined functions in Myria """
        return self._wrap_get('/function')

    def _get_plan(self, query, language, plan_type, **kwargs):
        catalog = MyriaCatalog(self)
        algebra = MyriaHyperCubeAlgebra(catalog) \
            if kwargs.get('multiway_jon', False) \
            else MyriaLeftDeepTreeAlgebra()

        if language.lower() == "datalog":
            return self._get_datalog_plan(query, plan_type, algebra, **kwargs)
        elif language.lower() in ["myrial", "sql"]:
            return self._get_myrial_or_sql_plan(query, plan_type,
                                                catalog, algebra, **kwargs)
        else:
            raise NotImplementedError('Language %s not supported' % language)

    def _get_myrial_or_sql_plan(self, query, plan_type, catalog, algebra,
                                **kwargs):
        parsed = Parser().parse(query, udas=[(udf['name'], udf['outputType'])
                                             for udf in self._get_udfs()])
        processor = interpreter.StatementProcessor(catalog)
        processor.evaluate(parsed)

        if plan_type == 'logical':
            return processor.get_physical_plan(
                target_alg=OptLogicalAlgebra())
        elif plan_type == 'physical':
            return processor.get_physical_plan(
                target_alg=algebra,
                multiway_join=kwargs.get('multiway_join', False),
                push_sql=kwargs.get('push_sql', True))
        else:
            raise NotImplementedError('Myria plan type %s' % plan_type)

    @staticmethod
    def _get_datalog_plan(query, plan_type, algebra, **kwargs):
        datalog = RACompiler()
        datalog.fromDatalog(query)

        if not datalog.logicalplan:
            raise SyntaxError("Unable to parse Datalog")
        elif plan_type == 'logical':
            return datalog.logicalplan
        elif plan_type == 'physical':
            datalog.optimize(target=algebra,
                             push_sql=kwargs.get('push_sql', True))
            return datalog.physicalplan
        else:
            raise NotImplementedError('Datalog plan type %s' % plan_type)

    def _get_udfs(self):
        if not self._udfs:
            self._udfs = [self.get_function(name)
                          for name in self.get_functions()]
        return self._udfs